/*
 * BarChartRenderer.java
 *
 * Created on 11. Mai 2007, 06:15
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package zugvis;

import at.ac.tuwien.cs.timevis.FisheyeTimeScale;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RectangularShape;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;

import prefuse.Constants;
import prefuse.render.AbstractShapeRenderer;
import prefuse.util.ColorLib;
import prefuse.util.FontLib;
import prefuse.util.GraphicsLib;
import prefuse.visual.VisualItem;


/**
 * This is the renderer for both bar chart visualizations.
 * @author Hansjoerg
 */
public class BarChartRenderer extends AbstractShapeRenderer {
    
    /**
     * Newline delimiter of text
     */
    protected String m_delim = "\n";
    
    /**
     * Horizontal border of barchart
     */
    protected int m_horizBorder = 2;
    
    /**
     * Vertical border of barchart
     */
    protected int m_vertBorder  = 0;
    
    /**
     * Barheight
     */
    protected int m_barHeight  = 9;
    
    /**
     * Flag if draw stations
     */
    protected boolean drawStations = true;
    
    /** The holder for the currently computed bounding box */
    protected RectangularShape m_bbox  = new Rectangle2D.Double();
    //protected Point2D m_pt = new Point2D.Double(); // temp point
    
    /**
     * Font of labels
     */
    protected Font    m_font; // temp font holder
    
    /**
     * Fieldname of trainnames
     */
    protected String m_trainNameField = "trainName";
    
    /**
     * Name of the train for label
     */
    protected String m_trainName; // label text
    
    /**
     * Dimension of the trainnamelabel
     */
    protected Dimension m_trainNameDim = new Dimension(); // text width / height
    
    /**
     * Fieldname of the traininfo
     */
    protected String m_trainInfoField = "trainInfo";
    
    /**
     * Text of traininfo for label
     */
    protected String m_trainInfo; // label text
    
    /**
     * Dimension of traininfolabel
     */
    protected Dimension m_trainInfoDim = new Dimension(); // text width / height
    
    /**
     * Fieldname of fromstation
     */
    protected String m_stationFromField = "fromName";
    
    /**
     * Stationname of fromstation fo label
     */
    protected String m_stationFrom; // label text
    
    /**
     * Dimension of stationfromlabel
     */
    protected Dimension m_stationFromDim = new Dimension(); // text width / height
    
    /**
     * Fieldname of tostation
     */
    protected String m_stationToField = "toName";
    
    /**
     * Stationname of tostation for label
     */
    protected String m_stationTo; // label text
    
    /**
     * Dimension of stationtolabel
     */
    protected Dimension m_stationToDim = new Dimension(); // text width / height
    
    /**
     * Timescale of visualisation for calculating pixelposition for a dimension
     */
    protected FisheyeTimeScale timeScale = null;
    
    
    /**
     * Create a new LabelRenderer. By default the field "label" is used
     * as the field name for looking up text, and no image is used.
     */
    public BarChartRenderer() {
    }
    
    /**
     * Create a new BarChartRenderer. Draws bars of connections and labels of the start- and endstation.
     * @param trainNameField Fieldname of trainnames
     * @param trainInfoField Fieldname of traininfos
     * @param stationFromField Fieldname of startstationname
     * @param stationToField Fieldname of endstationname
     */
    public BarChartRenderer(String trainNameField, String trainInfoField, String stationFromField, String stationToField) {
        
        m_trainNameField = trainNameField;
        m_trainInfoField = trainInfoField;
        m_stationFromField = stationFromField;
        m_stationToField = stationToField;
    }

    
    /**
     * Get the text of a field
     * @return The text of the field, or null if empty
     * @param fieldName Name of the field
     */
    public String getDataField(String fieldName) {
        
        if (fieldName.equals("trainName")) return m_trainNameField;
        if (fieldName.equals("trainInfo")) return m_trainInfoField;
        if (fieldName.equals("fromName")) return m_stationFromField;
        if (fieldName.equals("toName")) return m_stationToField;
        return null;
    }    
   
    
    /**
     * This sets the drawstations flag.
     * @param draw If true stations of a connection will be drawn as small white lines.
     */
    public void DrawStations(boolean draw){
        drawStations = draw;
    }
    
    
    /**
     * Returns the text to draw. Subclasses can override this class to
     * perform custom text selection.
     * @return a <code>String</code> to draw
     * @param dataField Text of a specific datafield
     * @param item the item to represent as a <code>String</code>
     */
    protected String getText(VisualItem item, String dataField) {
        String s = null;
        if ( item.canGetString(dataField) ) {
            return item.getString(dataField);
        }
        return s;
    }    
    
    
    // ------------------------------------------------------------------------
    // Rendering
    
    /**
     * Computes dimensions of labels trainname, statstationname and endstationname
     * @param item Visualitem to compute
     * @param size Fontsize
     */
    private void computeTextDimensions(VisualItem item, double size) {
        
        // put item font in temp member variable
        m_font = item.getFont();
        // scale the font as needed
        if ( size != 1 ) {
            m_font = FontLib.getFont(m_font.getName(), m_font.getStyle(),
                    size*m_font.getSize());
        }
        
        FontMetrics fm = DEFAULT_GRAPHICS.getFontMetrics(m_font);
        StringBuffer str = null;
        
        
        // trainName -----------------------------------------------------------
        
        // compute the number of lines and the maximum width
        m_trainNameDim.width = 0;
        m_trainNameDim.height = 0;
        
        if (m_trainName != null) {
            
            int nlines = 1, w = 0, start = 0, end = m_trainName.indexOf(m_delim);
            
            String line;
            for ( ; end >= 0; ++nlines ) {
                w = fm.stringWidth(line=m_trainName.substring(start,end));
                str.append(line).append(m_delim);
                m_trainNameDim.width = Math.max(m_trainNameDim.width, w);
                start = end+1;
                end = m_trainName.indexOf(m_delim, start);
            }
            w = fm.stringWidth(line=m_trainName.substring(start));
            
            if ( str != null ) str.append(line);
            
            m_trainNameDim.width = Math.max(m_trainNameDim.width, w); // update maximum width
            m_trainNameDim.height = fm.getHeight() * nlines;    // compute the text height
        }
        
        
//        // trainInfo -----------------------------------------------------------
//
//        // compute the number of lines and the maximum width
//        m_trainInfoDim.width = 0;
//        m_trainInfoDim.height = 0;
//
//        if (m_trainInfo != null) {
//
//            int nlines = 1, w = 0, start = 0, end = m_trainInfo.indexOf(m_delim);
//
//            String line;
//            for ( ; end >= 0; ++nlines ) {
//                w = fm.stringWidth(line=m_trainInfo.substring(start,end));
//                str.append(line).append(m_delim);
//                m_trainInfoDim.width = Math.max(m_trainInfoDim.width, w);
//                start = end+1;
//                end = m_trainInfo.indexOf(m_delim, start);
//            }
//            w = fm.stringWidth(line=m_trainInfo.substring(start));
//
//            if ( str != null ) str.append(line);
//
//            m_trainInfoDim.width = Math.max(m_trainInfoDim.width, w); // update maximum width
//            m_trainInfoDim.height = fm.getHeight() * nlines;    // compute the text height
//        }
        
        
        // stationFrom -----------------------------------------------------------
        
        // compute the number of lines and the maximum width
        m_stationFromDim.width = 0;
        m_stationFromDim.height = 0;
        
        if (m_stationFrom != null) {
            
            int nlines = 1, w = 0, start = 0, end = m_stationFrom.indexOf(m_delim);
            
            String line;
            for ( ; end >= 0; ++nlines ) {
                w = fm.stringWidth(line=m_stationFrom.substring(start,end));
                str.append(line).append(m_delim);
                m_stationFromDim.width = Math.max(m_stationFromDim.width, w);
                start = end+1;
                end = m_stationFrom.indexOf(m_delim, start);
            }
            w = fm.stringWidth(line=m_stationFrom.substring(start));
            
            if ( str != null ) str.append(line);
            
            m_stationFromDim.width = Math.max(m_stationFromDim.width, w); // update maximum width
            m_stationFromDim.height = fm.getHeight() * nlines;    // compute the text height
        }
        
        
        // stationTo -----------------------------------------------------------
        
        // compute the number of lines and the maximum width
        m_stationToDim.width = 0;
        m_stationToDim.height = 0;
        
        if (m_stationTo != null) {
            
            int nlines = 1, w = 0, start = 0, end = m_stationTo.indexOf(m_delim);
            
            String line;
            for ( ; end >= 0; ++nlines ) {
                w = fm.stringWidth(line=m_stationTo.substring(start,end));
                str.append(line).append(m_delim);
                m_stationToDim.width = Math.max(m_stationToDim.width, w);
                start = end+1;
                end = m_stationTo.indexOf(m_delim, start);
            }
            w = fm.stringWidth(line=m_stationTo.substring(start));
            
            if ( str != null ) str.append(line);
            
            m_stationToDim.width = Math.max(m_stationToDim.width, w); // update maximum width
            m_stationToDim.height = fm.getHeight() * nlines;    // compute the text height
        }
    }
    
    
    /**
     * Compute the boundingbos of the bar including textlabels
     * @param item Visualitem to compute
     * @return the shape with setted dimensions
     */
    protected Shape getRawShape(VisualItem item) {
        
        double x = item.getX();
        if ( Double.isNaN(x) || Double.isInfinite(x) ) x = 0;
        
        double y = item.getY();
        if ( Double.isNaN(y) || Double.isInfinite(y) ) y = 0;
        
        double barW = timeScale.getPixelForDate(item.getDate("toTime").getTime()) - x;
        double barH = m_barHeight;
        
        m_trainName   = getText(item, m_trainNameField);
        m_trainInfo   = getText(item, m_trainInfoField);;
        m_stationFrom = getText(item, m_stationFromField);;
        m_stationTo   = getText(item, m_stationToField);;
        
        double size = item.getSize();
        
        // get text dimensions
        computeTextDimensions(item, size);
        
        int tw=0, th=0;
        
        th = m_trainNameDim.height;
        tw = m_trainNameDim.width;        
        
        // render text only if smaller than the bar
        if (m_trainNameDim.width > barW) m_trainName = null;
        if (m_stationFromDim.width+10 > barW) m_stationFrom = null;
        if (m_stationToDim.width+10 > barW) m_stationTo = null;
        
        // get bounding box dimensions
        double w=0, h=0;
        w = barW;
        h = barH + m_stationFromDim.height + m_stationToDim.height;
        
        y -= h/2;
        
        m_bbox.setFrame(x, y, w, h);
        
        return m_bbox;
    }
    
    
    /**
     * Renders the barchart (including labels if enough place)
     * @param g Graphics2D
     * @param item Visualitem to draw
     */
    public void render(Graphics2D g, VisualItem item) {
        
        RectangularShape shape = (RectangularShape)getShape(item);
        RectangularShape shapeBar = (RectangularShape)getShape(item);
        
        if ( shape == null ) return;
        
        double size = item.getSize();
        
        shapeBar.setFrame(shapeBar.getX(), shapeBar.getY()+shape.getHeight()-m_barHeight,
                shapeBar.getWidth(), m_barHeight+2*size*m_horizBorder);
        
        // fill the shape, if requested
        int type = getRenderType(item);
        if ( type==RENDER_TYPE_FILL || type==RENDER_TYPE_DRAW_AND_FILL )
            GraphicsLib.paint(g, item, shapeBar, getStroke(item), RENDER_TYPE_FILL);
        
        // render stopOvers
        SimpleDateFormat dateFormat = new SimpleDateFormat( "dd.MM.yy" );
        SimpleDateFormat dateTimeFormat = new SimpleDateFormat( "dd.MM.yy HH:mm" );
        
        double shapeStopOverLine_Y1 = shapeBar.getY()+shape.getHeight()-m_barHeight;
        double shapeStopOverLine_Y2 = shapeBar.getY()+shape.getHeight();
        
        String stopOversString = item.getString("stopOvers");
        if (stopOversString.endsWith("~")) stopOversString = stopOversString.substring(0, stopOversString.length()-1);
        String stopOvers[] = stopOversString.split("~");        
        
        for (int i = 0; i < stopOvers.length; i++) {
            
            try { 
                
                Date toTime = dateTimeFormat.parse(
                        dateFormat.format(item.getDate("toTime").getTime())
                        + " " +
                        stopOvers[i]);
                
                
                double X = timeScale.getPixelForDate(toTime.getTime());
                
                g.setPaint(ColorLib.getColor(ColorLib.color(Color.WHITE)));
                //g.drawLine((int)X, (int)shapeStopOverLine_Y2-2, (int)X, (int)shapeStopOverLine_Y2-m_barHeight-3);

                if(drawStations) g.drawRect((int)X, (int)shapeStopOverLine_Y2-1, 1, 0);
                //g.drawRect((int)X, (int)shapeStopOverLine_Y2-m_barHeight-3, 1, 1);
                
            } catch (ParseException ex) {
                System.err.println("Can't combine \"stopOvers\"");
            }

        }

        
        // now render the text  
        boolean useInt = 1.5 > Math.max(g.getTransform().getScaleX(),
                g.getTransform().getScaleY());  
        
        // render text
        int textColor = item.getTextColor();
        if (ColorLib.alpha(textColor) > 0 ) {
            
            g.setFont(m_font);
            FontMetrics fm = DEFAULT_GRAPHICS.getFontMetrics(m_font);
            
            double x = shape.getMinX() + size*m_horizBorder;
            double y = shape.getMinY() + size*m_vertBorder;
            double w = shape.getWidth() - 2*size*m_horizBorder;
            double h = shape.getHeight() - 2*size*m_vertBorder;
            
            y += fm.getAscent();
            
            if ( m_trainName != null) {
                
                double yTemp = y + (h - m_trainNameDim.height);
                g.setPaint(ColorLib.getColor(textColor));
                drawStringLines(g, useInt, fm, m_trainName,
                        x, yTemp, w, Constants.CENTER, ColorLib.getColor(item.getStrokeColor()));
            }
            
            if ( m_stationFrom!= null) {
                
                double yTemp = y - m_stationFromDim.height;;// + (h - m_stationFromDim.height);  // for vertical center
                g.setPaint(ColorLib.getColor(ColorLib.color(Color.BLACK)));
                drawStringLines(g, useInt, fm, m_stationFrom,
                        x, yTemp, w, Constants.LEFT, null);
            }
            
            if ( m_stationTo!= null) {
                
                double yTemp = y + h;// - m_stationToDim.height;// + (h - m_stationFromDim.height);  // for vertical center
                g.setPaint(ColorLib.getColor(ColorLib.color(Color.BLACK)));
                drawStringLines(g, useInt, fm, m_stationTo,
                        x, yTemp, w, Constants.RIGHT, null);
            }
        }
        
        // draw border
        if (type==RENDER_TYPE_DRAW || type==RENDER_TYPE_DRAW_AND_FILL) {
            GraphicsLib.paint(g,item,shapeBar,getStroke(item),RENDER_TYPE_DRAW);
        }
    }
    
    
    /**
     * Reders lines of text
     * @param g Graphiks2D
     * @param useInt Round position to integer
     * @param fm FontMetrics
     * @param text Text to render
     * @param x Position x
     * @param y Position y
     * @param w Witdh of label (for center and right align)
     * @param align Textalign
     * @param bgColor Backgroudcolor
     */
    private final void drawStringLines(Graphics2D g, boolean useInt, FontMetrics fm,
            String text, double x, double y, double w, int align, Color bgColor) {
        
        // render each line of text
        int lh = fm.getHeight(); // the line height
        int start = 0, end = text.indexOf(m_delim);
        for ( ; end >= 0; y += lh ) {
            drawString(g, fm, text.substring(start, end), useInt, x, y, w, align, bgColor);
            start = end+1;
            end = text.indexOf(m_delim, start);
        }
        drawString(g, fm, text.substring(start), useInt, x, y, w, align, bgColor);
    }
    
    
    /**
     * Reders a line of text
     * @param g Graphiks2D
     * @param useInt Round position to integer
     * @param fm FontMetrics
     * @param text Text to render
     * @param x Position x
     * @param y Position y
     * @param w Witdh of label (for center and right align)
     * @param align Textalign
     * @param bgColor Backgroudcolor
     */
    private final void drawString(Graphics2D g, FontMetrics fm, String text,
            boolean useInt, double x, double y, double w, int align, Color bgColor) {
        // compute the x-coordinate
        double tx;
        switch ( align ) {
            case Constants.LEFT:
                tx = x;
                break;
            case Constants.RIGHT:
                tx = x + w - fm.stringWidth(text);
                break;
            case Constants.CENTER:
                tx = x + (w - fm.stringWidth(text)) / 2;
                break;
            default:
                throw new IllegalStateException(
                        "Unrecognized text alignment setting.");
        }
        // use integer precision unless zoomed-in
        // results in more stable drawing
        if (bgColor != null) g.setBackground(bgColor);
        if ( useInt ) {
            g.drawString(text, (int)tx, (int)y);            
        } else {
            g.drawString(text, (float)tx, (float)y);
        }
    }        
    
    // ------------------------------------------------------------------------
    
    
    /**
     * Returns the amount of padding in pixels between the content
     * and the border of this item along the horizontal dimension.
     * @return the horizontal padding
     */
    public int getHorizontalPadding() {
        return m_horizBorder;
    }
    

    /**
     * Sets the amount of padding in pixels between the content
     * and the border of this item along the horizontal dimension.
     * @param xpad the horizontal padding to set
     */
    public void setHorizontalPadding(int xpad) {
        m_horizBorder = xpad;
    }
    
    
    /**
     * Returns the amount of padding in pixels between the content
     * and the border of this item along the vertical dimension.
     * @return the vertical padding
     */
    public int getVerticalPadding() {
        return m_vertBorder;
    }
    
    
    /**
     * Sets the amount of padding in pixels between the content
     * and the border of this item along the vertical dimension.
     * @param ypad the vertical padding
     */
    public void setVerticalPadding(int ypad) {
        m_vertBorder = ypad;
    }
    
    
    /**
     * Sets the timeScale
     * @param _timeScale the timeScale
     */
    public void setTimeScale(FisheyeTimeScale _timeScale) {
        
        timeScale = _timeScale;
    }
    
    
} // end of class LabelRenderer
