/*
 * BarChartVis.java
 *
 * Created on 3. Juni 2007, 15:13
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package zugvis;

import at.ac.tuwien.cs.timevis.BasicTimeScale;
import at.ac.tuwien.cs.timevis.FisheyeTimeScale;
import at.ac.tuwien.cs.timevis.prefuse.TimeLayout;
import at.ac.tuwien.cs.timevis.ui.TimeScalePainter;
import java.awt.AlphaComposite;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.Point2D;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.Vector;
import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JToolTip;
import javax.swing.SwingUtilities;
import javax.swing.border.Border;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import prefuse.Constants;
import prefuse.Display;
import prefuse.Visualization;
import prefuse.action.ActionList;
import prefuse.action.GroupAction;
import prefuse.action.RepaintAction;
import prefuse.action.assignment.ColorAction;
import prefuse.action.layout.AxisLayout;
import prefuse.controls.ControlAdapter;
import prefuse.controls.PanControl;
import prefuse.controls.ZoomControl;
import prefuse.data.Table;
import prefuse.render.DefaultRendererFactory;
import prefuse.util.ColorLib;
import prefuse.util.display.PaintListener;
import prefuse.util.ui.JRangeSlider;
import prefuse.util.ui.ValuedRangeModel;
import prefuse.visual.VisualItem;
import prefuse.visual.VisualTable;

/**
 * This is the abstract class that holds most functionality for both BarChart-visualizations.
 * @author hansjoerg
 */
public abstract class BarChartVis extends JPanel {
    
    /**
     * Table of data
     */
    static Table data = null;
    /**
     * Vector of selected trains
     */
    static Vector selectedTrains = null;
    /**
     * Flag to display directions (both, from-to, to-from)
     */
    static int direction = 0;
    /**
     * Dispaly of visualization
     */
    static BarChartDisplay barChartDisplay = null;

    
    /** Creates a new instance of BarChartVis */
    public BarChartVis() {
        
        setLayout(new BorderLayout());
    }
    
    /**
     * Initializes the rangeslider of y-axis
     * @param barChartDisplay Display of visualization
     * @return the initialized rangeslider
     */
    static JRangeSlider makeJRangeSlider_Y(final BarChartDisplay barChartDisplay) {
        
        final JRangeSlider jRangeSlider_Y = new JRangeSlider(
                barChartDisplay.axisLayout_Y.getRangeModel().getMinimum(),
                barChartDisplay.axisLayout_Y.getRangeModel().getMaximum(),
                barChartDisplay.axisLayout_Y.getRangeModel().getMinimum(),
                barChartDisplay.axisLayout_Y.getRangeModel().getMaximum()-barChartDisplay.axisLayout_Y.getRangeModel().getMinimum(),
                JRangeSlider.VERTICAL, 1);
        
        jRangeSlider_Y.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent evt) {
                barChartDisplay.axisLayout_Y.getRangeModel().setValue(jRangeSlider_Y.getLowValue());
                barChartDisplay.axisLayout_Y.getRangeModel().setExtent(jRangeSlider_Y.getHighValue()-jRangeSlider_Y.getLowValue());
                barChartDisplay.getVisualization().run("repaint");
            }
        });
        
        return jRangeSlider_Y;
    }
    
    
    /**
     * The display of the barchartvisualisation
     */
    static class BarChartDisplay extends Display {
        
        /**
         * Name of internal datatable
         */
        private static final String DATA = "data";
        
        /**
         * Timescale of the visualization
         */
        FisheyeTimeScale timeScale = null;
        /**
         * The timescalepainter
         */
        private TimeScalePainter timeScalePainter = null;
        /**
         * The mousetracker
         */
        MouseTrackerBarChart mouseTracker = null;
        /**
         * The layout of the y-axis
         */
        AxisLayout axisLayout_Y = null;
        /**
         * The renderer for the bars
         */
        BarChartRenderer barRenderer = null;
        
        
        /*
         * Zoom & Pan: Instead of changing the Display's AffineTransform, just
         * change the TimeScale here so we can use the built-in prefuse Zoom & Pan
         * Controls. Note: Changing of the Display's AffineTransfrom ist not
         * supported by timevis (except y-panning)!
         */
        /**
         * Perform panaction
         * @param dx Difference of x-axis
         * @param dy Difference of y-axis
         */
        public synchronized void pan(double dx, double dy) {
            panAbs(dx, dy);
        }
        /**
         * Perform panaction
         * @param dx Difference of x-axis
         * @param dy Difference of y-axis
         */        
        public synchronized void panAbs(double dx, double dy) {
            if (timeScale != null) {
                
                long oldStartDate = timeScale.getStartDate();
                long oldEndDate = timeScale.getEndDate();
                
                long diffMillis = timeScale.getMillisPerPixel() * (int)dx;
                
                long newStartDate = oldStartDate - diffMillis;
                long newEndDate = oldEndDate - diffMillis;
                
                timeScale.setStartDate(newStartDate);
                
                barChartDisplay.getParent().getParent().getParent().firePropertyChange(
                        "_STARTDATE_", oldStartDate, newStartDate
                        );
                
                timeScale.setEndDate(newEndDate);
                
                barChartDisplay.getParent().getParent().getParent().firePropertyChange(
                        "_ENDDATE_", oldEndDate, newEndDate
                        );
            }
        }
        /**
         * Perform zoomaction
         * @param p Point2D (unused)
         * @param scale zoomfactor
         */         
        public synchronized void zoom(Point2D p, double scale) {
            zoomAbs(p, scale);
        }
        
        /**
         * Perform zoomaction
         * @param p Point2D (unused)
         * @param scale zoomfactor
         */         
        public synchronized void zoomAbs(Point2D p, double scale) {
            if (timeScale != null) {
                
                Date oldStartDate = new Date(timeScale.getStartDate());
                Date oldEndDate = new Date(timeScale.getEndDate());
                
                timeScale.zoom(scale);
                
                Date newStartDate = new Date(timeScale.getStartDate());
                Date newEndDate = new Date(timeScale.getEndDate());
                
                barChartDisplay.getParent().getParent().getParent().firePropertyChange(
                        "_STARTDATE_",
                        oldStartDate.getTime(),
                        newStartDate.getTime()
                        );
                barChartDisplay.getParent().getParent().getParent().firePropertyChange(
                        "_ENDDATE_",
                        oldEndDate.getTime(),
                        newEndDate.getTime()
                        );
            }
        }
        
        /* End of Zoom & Pan */
        
        
        /**
         * Creates a new tooltip
         * @return new tooltip
         */
        public JToolTip createToolTip() {
            
            JToolTip jToolTip = new JToolTip();
            //jToolTip.setForeground(Color.BLUE);
            jToolTip.setBackground(Color.LIGHT_GRAY);
            //jToolTip.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
            
            return jToolTip;
        }
        
        /**
         * The display of the visualization
         * @param data Table of data
         * @param enableZoomAndPan Flag if zomm and pan is enabled
         */
        public BarChartDisplay(Table data, boolean enableZoomAndPan) {
            super(new Visualization());
            
            setSize(new Dimension(700, 400));
            setDamageRedraw(false);
            GregorianCalendar d = new GregorianCalendar();
            
            timeScale = new FisheyeTimeScale();
            
            timeScale.setAdjustDateRangeOnResize(false);
            timeScale.addChangeListener(new ChangeListener() {
                public void stateChanged(ChangeEvent e) {
                    getVisualization().run("layout");
                }
            });
            
            timeScalePainter = new TimeScalePainter(this);
            timeScalePainter.setTimeScale(timeScale);
            
            DefaultRendererFactory rf = new DefaultRendererFactory();
            
            barRenderer = new BarChartRenderer();
            barRenderer.setTimeScale(timeScale);
            
            rf.setDefaultRenderer(barRenderer);
            getVisualization().setRendererFactory(rf);
            
            addPaintListener(new PaintListener() {
                public void postPaint(Display d, Graphics2D g) {
                    mouseTracker.paintTimeAtPosition(g);
                }
                
                public void prePaint(Display d, Graphics2D g) {
                    g.transform(d.getInverseTransform());
                    timeScalePainter.paint(g);
                    g.transform(d.getTransform());
                }
            });
            
            if (enableZoomAndPan) {
                
                addControlListener(new PanControl());
                addControlListener(new ZoomControl());
            }
            
            addControlListener(new BarChartControl());
            
            addComponentListener(new ComponentAdapter() {
                public void componentResized(ComponentEvent e) {
                    timeScale.setDisplayWidth(getWidth() - 1);
                }
            });
            
            getVisualization().addTable(DATA, data);
            
            axisLayout_Y = new AxisLayout(DATA, "internalConnectionID", Constants.Y_AXIS);
            axisLayout_Y.setRangeModel(new BarChartValuedRangeModel(data, "internalConnectionID"));
            
            ActionList layout = new ActionList();
            layout.add(axisLayout_Y);
            layout.add(new TimeLayout(DATA, "fromTime", timeScale));
            
            layout.add(new ColorAction(DATA, VisualItem.TEXTCOLOR, ColorLib.color(Color.WHITE)));
            layout.add(new ColorAction(DATA, VisualItem.FILLCOLOR, ColorLib.rgb(50,50,125))); //ColorLib.color(Color.LIGHT_GRAY)));
            //layout.add(new ColorAction(DATA, VisualItem.STROKECOLOR, ColorLib.color(Color.BLUE.darker())));
            layout.add(new BarChartFilter(getVisualization(), DATA));
            layout.add(new RepaintAction());
            getVisualization().putAction("layout", layout);
            
            ActionList repaint = new ActionList();
            repaint.add(axisLayout_Y);
            repaint.add(new BarChartFilter(getVisualization(), DATA));
            repaint.add(new RepaintAction());
            getVisualization().putAction("repaint", repaint);
            
            getVisualization().run("layout");
        }
        
    }
    
    
    
    /**
     * Filter of connections
     */
    static class BarChartFilter extends GroupAction {
        
        /**
         * Is item visible
         */
        private boolean visible = true;
        /**
         * Visualization that contains this item
         */
        private Visualization vis = null;
        
        /**
         * Create a new instance of the filter
         * @param vis Visualization that contains this item
         * @param group Grupname of visualitem
         */
        public BarChartFilter(Visualization vis, String group){
            super(group);
            this.vis = vis;
        }
        
        /**
         * Runs the filter
         * @param frac (unused)
         */
        public void run(double frac){
            
            Iterator items = ((VisualTable)vis.getGroup("data")).tuples();

            while( items.hasNext()){
                
                VisualItem item = (VisualItem)items.next();

                // color selected trains
                if (selectedTrains.contains(item.getString("trainID"))) 
                    item.setFillColor(ColorLib.rgb(250, 0, 0));
                else item.setFillColor(ColorLib.rgb(50,50,125));                
                
                // display only trains in direction
                int from = item.getInt("fromKm");
                int to = item.getInt("toKm");
                
                if(from > to && item.getFillColor() == ColorLib.rgb(50, 50, 125)){
                    item.setFillColor(ColorLib.rgb(120, 120, 255));
                    item.setEndTextColor(ColorLib.rgb(120, 120, 255));
                }
                
                switch(direction){
                    case 1: 
                        item.setVisible(from < to); 
                        break;  // UP                      
                    case 2: 
                        item.setVisible(from > to); 
                        break;  // DOWN                      
                    default: item.setVisible(true); break;      // UPDOWN
                }                
            }
        }
    }
         
   
    /**
     * Rangemodel for y-axis
     */
    static class BarChartValuedRangeModel implements ValuedRangeModel {
        
        private Table data;
        private int minimum, maximum, value, extent;
        private boolean adjusting;
        private Vector changeListener = new Vector();
        
        /**
         * Creates a new instance of the rangemodel
         * @param data Table of data
         * @param row Rowname of internalConnectionID
         */
        public BarChartValuedRangeModel(Table data, String row) {
            
            this.data = data;
            minimum = value = data.getInt(0, row)-15;
            maximum = extent = data.getInt(data.getRowCount()-1, row)+7;
            adjusting = false;
        }
        
        public Object getMinValue() {
            return minimum;
            // id
        }
        
        public Object getMaxValue() {
            return maximum;
            // id
        }
        
        public Object getLowValue() {
            return value;
            // id
        }
        
        public Object getHighValue() {
            return value + extent;
            // id
        }
        
        public int getMinimum() { return minimum; }
        
        public void setMinimum(int newMinimum) { minimum = newMinimum;  }
        
        public int getMaximum() { return maximum; }
        
        public void setMaximum(int newMaximum) { maximum = newMaximum; }
        
        public int getValue() { return value; }
        
        public void setValue(int newValue) { value = newValue; }
        
        public void setValueIsAdjusting(boolean b) { adjusting = b; }
        
        public boolean getValueIsAdjusting() { return adjusting; }
        
        public int getExtent() { return extent; }
        
        public void setExtent(int newExtent) { extent = newExtent; }
        
        public void setRangeProperties(int value, int extent, int minimum, int maximum, boolean adjusting) {
            
            this.value = value;
            this.extent = extent;
            this.minimum = minimum;
            this.maximum = maximum;
            this.adjusting = adjusting;
        }
        
        public void addChangeListener(ChangeListener x) { changeListener.add(x); }
        
        public void removeChangeListener(ChangeListener x) { changeListener.remove(x); }
        
    }
    
    
    
    static class BarChartControl extends ControlAdapter {
        
        private VisualItem activeItem;
        private Point2D down = new Point2D.Double();
        private Point2D tmp = new Point2D.Double();
        private boolean wasFixed, dragged;
        private boolean repaint = false;
        
        public void itemEntered(VisualItem item, MouseEvent e) {
            
            ((Display)e.getSource()).setToolTipText(item.getString("tipText"));
            item.setEndTextColor(item.getFillColor());
            item.setFillColor(ColorLib.rgb(250,100,0));
        }
        
        public void itemExited(VisualItem item, MouseEvent e) {

            Display d = (Display)e.getSource();
            d.setToolTipText(null);
            item.setFillColor(item.getEndTextColor());
        }
//
//        public void itemPressed(VisualItem item, MouseEvent e) {
//            if (!SwingUtilities.isLeftMouseButton(e)) return;
//
//            // set the focus to the current node
//            Visualization vis = item.getVisualization();
//            vis.getFocusGroup(Visualization.FOCUS_ITEMS).setTuple(item);
//
//            item.setFixed(true);
//            dragged = false;
//            Display d = (Display)e.getComponent();
//            down = d.getAbsoluteCoordinate(e.getPoint(), down);
//
//            vis.run("forces");
//        }
//
//        public void itemReleased(VisualItem item, MouseEvent e) {
//            if (!SwingUtilities.isLeftMouseButton(e)) return;
//            if ( dragged ) {
//                activeItem = null;
//                item.setFixed(wasFixed);
//                dragged = false;
//            }
//            // clear the focus
//            Visualization vis = item.getVisualization();
//            vis.getFocusGroup(Visualization.FOCUS_ITEMS).clear();
//
//            vis.cancel("forces");
//        }
//
        public void itemClicked(VisualItem item, MouseEvent e) {
            if (!SwingUtilities.isLeftMouseButton(e)) return;
            if ( e.getClickCount() == 1 ) {
                if (selectedTrains.contains(item.getString("trainID"))){
                    selectedTrains.remove(item.getString("trainID"));
                    item.setFillColor(ColorLib.rgb(50, 50, 125));
                    item.setEndTextColor(ColorLib.rgb(50, 50, 125));
                    
                }
                else{
                    selectedTrains.add(item.getString("trainID"));
                    item.setFillColor(ColorLib.rgb(255, 0, 0));
                    item.setEndTextColor(ColorLib.rgb(255, 0, 0));
                }
                //item.getVisualization().run("repaint");
                

                
                // fire PropertyChange on TabbedPane, the container of oll visualisations
                barChartDisplay.getParent().getParent().getParent().firePropertyChange("_REPAINT_", 0, 1);
            }
        }
//
//        public void itemDragged(VisualItem item, MouseEvent e) {
//            if (!SwingUtilities.isLeftMouseButton(e)) return;
//            dragged = true;
//            Display d = (Display)e.getComponent();
//            tmp = d.getAbsoluteCoordinate(e.getPoint(), tmp);
//            double dx = tmp.getX()-down.getX();
//            double dy = tmp.getY()-down.getY();
//
//            PrefuseLib.setX(item, null, item.getX()+dx);
//            PrefuseLib.setY(item, null, item.getY()+dy);
//            down.setLocation(tmp);
//            if ( repaint )
//                item.getVisualization().repaint();
//        }
    } // end of class BarChartControl
    
    /**
     * Tracks the mouse position on a given {@link JComponent}.
     * {@link MouseTracker#paintTimeAtPosition(Graphics2D)} paints a vertical line
     * with a text label reperesenting the TIME at the current mouse position. The
     * date is provided by {@link BasicTimeScale}.
     */
    static class MouseTrackerBarChart extends MouseMotionAdapter {
        
        private int lastX;
        private BasicTimeScale timeScale;
        private JComponent component;
        private boolean timeOnly;
        
        /**
         * Create a {@link MouseTracker} which tracks the mouse position on the
         * given component. The {@link BasicTimeScale} is used to get the date on a
         * given position.
         *
         * @param component
         *            the component
         */
        public MouseTrackerBarChart(JComponent component, BasicTimeScale timeScale, boolean timeOnly) {
            this.timeScale = timeScale;
            this.component = component;
            this.timeOnly = timeOnly;
            component.addMouseMotionListener(this);
        }
        
        /**
         * Paint a vertical line at the mouse position.
         *
         * Typically this method should be called with a graphics object provided by
         * the {@link JComponent#paint(java.awt.Graphics)} method of the tracked
         * component.
         *
         * @param g
         *            the graphics object used for painting the tracked component
         */
        public void paintTimeAtPosition(Graphics2D g) {
            if (timeScale == null || component == null) {
                return;
            }
            
            Graphics2D g2d = (Graphics2D) g;
            
            String dateText = null;
            if (timeOnly) dateText = new SimpleDateFormat( "HH:mm:ss" ).format(timeScale.getDateAtPixel(lastX));
            else dateText = timeScale.getTimeUnit().getFullFormat().format(timeScale.getDateAtPixel(lastX));
            
            int textWidth = g.getFontMetrics().stringWidth(dateText);
            g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f));
            
            
            g.setColor(Color.BLACK);
            g.drawLine(lastX, 0, lastX, component.getHeight());
            g2d.setComposite(AlphaComposite.Src);
            g.setColor(Color.GRAY);
            g.drawString(dateText, (lastX + 1) - Math.max(0, ((lastX + textWidth) - component.getWidth())), component
                    .getHeight()
                    - g.getFontMetrics().getDescent());
        }
        
        /*
         * (non-Javadoc)
         *
         * @see java.awt.event.MouseMotionAdapter#mouseMoved(java.awt.event.MouseEvent)
         */
        public void mouseMoved(MouseEvent e) {
            if (component != null) {
                lastX = e.getX();
                component.repaint();
            }
        }
        
    }
    
    
}
