/*
 * BarChartVisAbs.java
 *
 * Created on 9. Mai 2007, 14:16
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package zugvis;

import at.ac.tuwien.cs.timevis.ui.StatusBar;
import at.ac.tuwien.cs.timevis.ui.TimeScaleHeader;
import java.awt.BorderLayout;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.GregorianCalendar;
import java.util.Vector;
import java.util.Iterator;
import javax.swing.Box;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import prefuse.data.Table;
import prefuse.visual.VisualItem;
import prefuse.visual.VisualTable;
import zugvis.BarChartVis.BarChartDisplay;
import zugvis.BarChartVis.MouseTrackerBarChart;


/**
 * This class contains the functionality for the timeline chart.
 * @author Hansjoerg
 */
public class BarChartVisAbs extends BarChartVis {
    
    /**
     * Creates a new instance of BarChartVisAbs
     * @param tabPane Tabpane that contains this visualisation (for propertychangeeventes)
     * @param selectedTrains Vector fo selected trains
     * @param connections Table of connections
     * @param stations Table of stations
     * @param trains Table of trains
     * @param dates Table of dates
     */
    public BarChartVisAbs(JTabbedPane tabPane, Vector selectedTrains, Table connections, Table stations, Table trains, Table dates) {
        
        super();
        
        this.selectedTrains = selectedTrains;
        
        connections = Functions.deleteStopOvers(connections);
        tabPane.firePropertyChange("_PUSHPROGRESS_", 0, 7);
        connections = Functions.getSortedConnections(connections, false);
        tabPane.firePropertyChange("_PUSHPROGRESS_", 0, 7);
        connections = Functions.addInternalConnectionIDs(connections);
        
        data = Functions.makeData(connections, stations, trains, dates);
        tabPane.firePropertyChange("_PUSHPROGRESS_", 0, 7);
        data = Functions.addTrainInfo(data, trains);
        data = Functions.addStationNames(data, stations);
        data = Functions.addTipText(data, dates, false);
        tabPane.firePropertyChange("_PUSHPROGRESS_", 0, 7);
        
        barChartDisplay = new BarChartDisplay(data, true);
        barChartDisplay.timeScale.setDisplayWidth(barChartDisplay.getWidth() - 1);
        barChartDisplay.timeScale.changeInterval(Functions.getEarliestDate(data, new GregorianCalendar()).getTime().getTime()-1000*60*30, Functions.getLatestDate(data, new GregorianCalendar()).getTime().getTime()+1000*60*30);
        barChartDisplay.mouseTracker = new MouseTrackerBarChart(barChartDisplay, barChartDisplay.timeScale, false);
        
        JPanel JPanelVis = new JPanel();
        JPanelVis.setLayout(new BorderLayout());
        JPanelVis.add(barChartDisplay, BorderLayout.CENTER);
        JPanelVis.add(makeJRangeSlider_Y(barChartDisplay), BorderLayout.EAST);
        JPanelVis.add(new TimeScaleHeader(barChartDisplay.timeScale), BorderLayout.NORTH);
        
        StatusBar statusBar = new StatusBar(barChartDisplay.timeScale);
        statusBar.add(Box.createHorizontalStrut(20));
        JPanelVis.add(statusBar, BorderLayout.SOUTH);
        
        add(JPanelVis, BorderLayout.CENTER);
        
        /****************************************************************************
         * This functions ware called for external events (menubar or other visualisatzions
         *
         *
         */
        addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                
                if (evt.getPropertyName().equals("_REPAINT_")) {
                    
                    barChartDisplay.getVisualization().run("repaint");
                }
                
                if (evt.getPropertyName().equals("_FISHEYE_")) { 
                    
                    double fisEye = (double)Functions.object2Long(evt.getNewValue());
                    barChartDisplay.timeScale.setFisheyeIntensity(fisEye/1000);
                }
                
                if (evt.getPropertyName().equals("_SETDIRECTION_")) { 
                    
                    direction = (int)Functions.object2Long(evt.getNewValue());
                    barChartDisplay.getVisualization().run("repaint");
                }       
                
                if (evt.getPropertyName().equals("_STARTDATE_")) {

                    long endDate = barChartDisplay.timeScale.getEndDate();
                    barChartDisplay.timeScale.changeInterval(
                            Functions.object2Long(evt.getNewValue()), 
                            endDate
                            );
                }
                
                if (evt.getPropertyName().equals("_ENDDATE_")) { 

                    long startDate = barChartDisplay.timeScale.getStartDate();
                    barChartDisplay.timeScale.changeInterval(
                            startDate,
                            Functions.object2Long(evt.getNewValue())                            
                            );                    
                }
                
                if (evt.getPropertyName().equals("_SHOWSTATIONS_")) { 
                    
                    long newVal = Functions.object2Long(evt.getNewValue());
                    barChartDisplay.barRenderer.DrawStations(newVal == 1 ? true : false);
                    barChartDisplay.getVisualization().run("repaint");
                }
                
                if (evt.getPropertyName().equals("_FITTIME_")){

                    boolean fisheyeEnabled = barChartDisplay.timeScale.isFisheyeEnabled();
                    if (fisheyeEnabled) barChartDisplay.timeScale.setFisheyeEnabled(false);

                    GregorianCalendar startDate = new GregorianCalendar();
                    startDate.setTimeInMillis(barChartDisplay.timeScale.getStartDate());

                    GregorianCalendar endDate = new GregorianCalendar();
                    endDate.setTimeInMillis(barChartDisplay.timeScale.getEndDate());
                    
                    GregorianCalendar newEndDate = new GregorianCalendar();
                    newEndDate.setTimeInMillis(startDate.getTimeInMillis());
                    
                    GregorianCalendar newStartDate = new GregorianCalendar();
                    newStartDate.setTimeInMillis(endDate.getTimeInMillis());

                    Iterator it = ((VisualTable)barChartDisplay.getVisualization().getGroup("data")).tuples();
                    
                    while(it.hasNext()){
                        VisualItem act = (VisualItem) it.next();
                        
                        if(act.getDate("toTime").getTime() < endDate.getTimeInMillis() && act.getDate("toTime").getTime() > newEndDate.getTimeInMillis())
                            newEndDate.setTimeInMillis(act.getDate("toTime").getTime() + 1);
                        
                        if(act.getDate("fromTime").getTime() < newStartDate.getTimeInMillis() && act.getDate("fromTime").getTime() > startDate.getTimeInMillis())
                            newStartDate.setTimeInMillis(act.getDate("fromTime").getTime() - 1);
                    }

                    //barChartDisplay.getParent().getParent().getParent().firePropertyChange("_STARTDATE_", 0, Functions.getEarliestDate((Table) barChartDisplay.getVisualization().getGroup("data"), startDate).getTime().getTime());
                    //barChartDisplay.getParent().getParent().getParent().firePropertyChange("_ENDDATE_", 0, Functions.getLatestDate((Table) barChartDisplay.getVisualization().getGroup("data"), endDate).getTime().getTime());
                    barChartDisplay.getParent().getParent().getParent().firePropertyChange("_STARTDATE_", 0, newStartDate.getTimeInMillis());
                    barChartDisplay.getParent().getParent().getParent().firePropertyChange("_ENDDATE_", 0, newEndDate.getTimeInMillis());

                    barChartDisplay.timeScale.setFisheyeEnabled(fisheyeEnabled);
                }
            }
            
        });
    }    
}
