/*
 * Functions.java
 *
 * Created on 10. Mai 2007, 14:35
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package zugvis;

import com.toedter.calendar.JCalendar;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Enumeration;
import java.util.GregorianCalendar;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;
import prefuse.Visualization;
import prefuse.data.Table;
import prefuse.data.Tuple;
import java.util.Iterator;
import prefuse.data.io.DelimitedTextTableReader;


/**
 * Set of utility functions
 * @author Hansjoerg Pirpamer, Georg Prohaska
 */
public class Functions {
    
    /**
     * Delete Stop overs
     * @param connections Tabel of Connections
     * @return Table of Connections without stop overs
     */
    public static Table deleteStopOvers(Table connections) {
        
        Table connections_new = new Table();
        
        // copy columns
        for (int i = 0; i < connections.getColumnCount(); i++)
            connections_new.addColumn(connections.getColumnName(i), connections.getColumnType(i));
        
        connections_new.addColumn("stopOvers", String.class);
        
        // add rows
        for (int i = 0; i < connections.getRowCount(); i++) {
            
            if (connections_new.getRowCount() == 0 ||
                    connections.getInt(i, "connectionID") != connections_new.getInt(connections_new.getRowCount()-1, "connectionID") ||
                    connections.getInt(i, "trainID") != connections_new.getInt(connections_new.getRowCount()-1, "trainID")) {
                
                connections_new.addTuple(connections.getTuple(i));
            }
            
            // Targetstation
            if (i+1 == connections.getRowCount() ||
                    (i+1 < connections.getRowCount() &&
                    (connections.getInt(i+1, "connectionID") != connections_new.getInt(connections_new.getRowCount()-1, "connectionID") ||
                    connections.getInt(i+1, "trainID") != connections_new.getInt(connections_new.getRowCount()-1, "trainID")))) {
                
                connections_new.setString(connections_new.getRowCount()-1, "to", connections.getString(i, "to"));
                connections_new.setString(connections_new.getRowCount()-1, "toTime", connections.getString(i, "toTime"));
            }
            // else StopOver
            else {
                
                String stopOverString = connections_new.getString(connections_new.getRowCount()-1, "stopOvers");
                
                if (stopOverString == null) stopOverString = "";
                
                stopOverString += connections.getString(i, "toTime") + "~";
                
                connections_new.setString(connections_new.getRowCount()-1, "stopOvers", stopOverString);
            }
        }
        
        return connections_new;
    }
    
    
    
    /**
     * Sort connectionstable on field "fromTime"
     * @return Sortet table of Connections
     * @param ascend Sort ascend
     * @param connections Tabel of Connections
     */
    public static Table getSortedConnections(Table connections, boolean ascend) {
        
        Table connections_new = new Table();
        connections.addColumn("formTimeInMillis", long.class);
        
        // copy columns
        for (int i = 0; i < connections.getColumnCount(); i++)
            connections_new.addColumn(connections.getColumnName(i), connections.getColumnType(i));
        
        
        GregorianCalendar tempD = new GregorianCalendar();
        tempD.set(tempD.get(tempD.YEAR), tempD.get(tempD.MONTH), tempD.get(tempD.DAY_OF_MONTH), 0, 0, 0);
        
        SimpleDateFormat dateFormat = new SimpleDateFormat( "dd.MM.yy" );
        SimpleDateFormat dateTimeFormat = new SimpleDateFormat( "dd.MM.yy HH:mm" );
        
        // add a temporaly row fromTimeMillis with fromTime in milliseconds
        for (int i = 0; i < connections.getRowCount(); i++) {
            try {
                
                connections.setLong(i, "formTimeInMillis",
                        dateTimeFormat.parse(
                        dateFormat.format(tempD.getTime())
                        + " " +
                        connections.getString(i, "toTime")).getTime());
            } catch (ParseException ex) {
                System.err.println("Can't read \"toTime\" form connections");
            }
        }
        
        // make a list with connectionIDs in ascent order of fromTimeInMillis
        Vector connectionIDs = new Vector();
        
        for (int i = 0; i < connections.getRowCount(); i++) {
            
            long maxDate = -1;
            String maxConnectionID = "";
            
            for (int j = 0; j < connections.getRowCount(); j++) {
                
                if (maxDate < connections.getLong(i, "formTimeInMillis") &&
                        !connectionIDs.contains(connections.getString(i, "connectionID"))) {
                    
                    maxDate = connections.getLong(i, "formTimeInMillis");
                    maxConnectionID = connections.getString(i, "connectionID");
                }
            }
            
            if (!maxConnectionID.equals("")) {
                connectionIDs.add(maxConnectionID);
            }
            
        }
        
        // reverse if !ascend
        if (!ascend) {
            
            Vector new_connectionIDs = new Vector();
            
            for (int i = connectionIDs.size()-1; i >= 0; i-- )
                new_connectionIDs.add(connectionIDs.get(i));
            
            
            connectionIDs = new_connectionIDs;
        }
        
        // add rows
        for (int i = 0; i < connectionIDs.size(); i++)
            for (int j = 0; j < connections.getRowCount(); j++)
                if ( ((String)connectionIDs.get(i)).equals(connections.getString(j, "connectionID")) )
                    connections_new.addTuple(connections.getTuple(j));
        
        // delete temporaly rows
        connections.removeColumn("formTimeInMillis");
        connections_new.removeColumn("formTimeInMillis");
        
        return connections_new;
    }
    
    
    /**
     * Geneartes a Table with all Dates of the planevalidation and trains
     * @param connections Table of connections
     * @param stations Table of stationnames and infos
     * @param trains Table of trainnames and infos
     * @param dates Table of dates, if a connection goes or not on a specific day/date (first row are the planvaluerange)
     * @return Table of all connections and infos
     */
    public static Table makeData(Table connections, Table stations, Table trains, Table dates) {
        
        Table data = new Table();
        data.addColumn("ID", double.class);
        data.addColumn("internalConnectionID", int.class);
        data.addColumn("connectionID", String.class);
        data.addColumn("from", int.class);
        data.addColumn("to", int.class);
        data.addColumn("toTime", Date.class);
        data.addColumn("fromTime", Date.class);
        data.addColumn("stopOvers", String.class);
        data.addColumn("trainID", String.class);
        
        
        SimpleDateFormat dateFormat = new SimpleDateFormat( "dd.MM.yy" );
        SimpleDateFormat dateTimeFormat = new SimpleDateFormat( "dd.MM.yy HH:mm" );
        
        GregorianCalendar fromDate = new GregorianCalendar();
        GregorianCalendar toDate = new GregorianCalendar();
        
        // read first line of dates
        try{
            
            fromDate.setTime(dateFormat.parse(dates.getString(0, "specialDate")));
            toDate.setTime(dateFormat.parse(dates.getString(0, "notOnDate")));
        } catch (Exception e) { System.err.println("Can't read times form dates"); }
        
        // security-counter to prevent a endless while
        int secCounter = 365*100;
        double IDcounter = 0;
        
        while (fromDate.compareTo(toDate) <= 0 && secCounter > 0) {
            
            for (int i = 0; i < connections.getRowCount(); i++) {
                
                // continue if connection don't goes on this date
                if (!trainOnDate(connections.getInt(i, "connectionID"), fromDate, dates)) continue;
                
                data.addRow();
                data.setString(data.getRowCount()-1, "connectionID", connections.getString(i, "connectionID"));
                
                // add the internConnectionID if available
                if (connections.canGet("internalConnectionID", int.class))
                    data.setInt(data.getRowCount()-1, "internalConnectionID", connections.getInt(i, "internalConnectionID"));
                
                // add the stopOvers if available
                if (connections.canGet("stopOvers", String.class))
                    data.setString(data.getRowCount()-1, "stopOvers", connections.getString(i, "stopOvers"));
                
                
                data.setInt(data.getRowCount()-1, "from", connections.getInt(i, "from"));
                data.setInt(data.getRowCount()-1, "to", connections.getInt(i, "to"));
                
                // read toTime
                GregorianCalendar toTime = new GregorianCalendar();
                toTime.setTime(fromDate.getTime());
                
                try {
                    
                    Date readTime = dateTimeFormat.parse(
                            dateFormat.format(toTime.getTime())
                            + " " +
                            connections.getString(i, "toTime"));
                    
                    toTime.setTime(readTime);
                    
                } catch (ParseException ex) {
                    System.err.println("Can't read \"toTime\" form connections");
                }
                
                data.setDate(data.getRowCount()-1, "toTime", toTime.getTime());
                
                
                // readFromTime
                GregorianCalendar fromTime = new GregorianCalendar();
                fromTime.setTime(fromDate.getTime());
                
                try {
                    
                    Date readTime = dateTimeFormat.parse(
                            dateFormat.format(fromTime.getTime())
                            + " " +
                            connections.getString(i, "fromTime"));
                    
                    fromTime.setTime(readTime);
                    
                } catch (ParseException ex) {
                    System.err.println("Can't read \"fromTime\" form connections");
                }
                
                data.setDate(data.getRowCount()-1, "fromTime", fromTime.getTime());
                
                data.setString(data.getRowCount()-1, "trainID", connections.getString(i, "trainID"));
                
                if(data.getRowCount() != 1)
                    if(!data.getString(data.getRowCount()-2, "trainID").equals(data.getString(data.getRowCount()-1, "trainID")) ||
                        !data.getString(data.getRowCount()-2, "connectionID").equals(data.getString(data.getRowCount()-1, "connectionID")))
                        IDcounter++;
                
                data.setDouble(data.getRowCount()-1, "ID", IDcounter);
            }
            
            // next day
            fromDate.add(Calendar.DAY_OF_MONTH, 1);
            secCounter--;
        }
        
        return data;
    }
    
    
    
    /**
     * Geneartes a Table with all Dates of the planevalidation and
     * trains for the relative barchartvisualization
     * @param connections Table of connections
     * @param stations Table of stationnames and infos
     * @param trains Table of trainnames and infos
     * @param dates Table of dates, if a connection goes or not on a specific day/date (first row are the planvaluerange)
     * @return Table of all connections and infos
     */
    public static Table makeRelData(Table connections, Table stations, Table trains, Table dates) {
        
        Table data = new Table();
        data.addColumn("internalConnectionID", int.class);
        data.addColumn("connectionID", String.class);
        data.addColumn("from", int.class);
        data.addColumn("to", int.class);
        data.addColumn("toTime", Date.class);
        data.addColumn("toTimeReal", Date.class);
        data.addColumn("fromTime", Date.class);
        data.addColumn("fromTimeReal", Date.class);
        data.addColumn("stopOvers", String.class);
        data.addColumn("trainID", String.class);
        
        GregorianCalendar tempD = new GregorianCalendar();
        tempD.set(tempD.get(tempD.YEAR), tempD.get(tempD.MONTH), tempD.get(tempD.DAY_OF_MONTH), 0, 0, 0);
        
        SimpleDateFormat dateFormat = new SimpleDateFormat( "dd.MM.yy" );
        SimpleDateFormat timeFormat = new SimpleDateFormat( "HH:mm" );
        SimpleDateFormat dateTimeFormat = new SimpleDateFormat( "dd.MM.yy HH:mm" );
        
        for (int i = 0; i < connections.getRowCount(); i++) {
            
            data.addRow();
            
            // add the internConnectionID if available
            if (connections.canGet("internalConnectionID", int.class))
                data.setInt(data.getRowCount()-1, "internalConnectionID", connections.getInt(i, "internalConnectionID"));
            
            data.setString(data.getRowCount()-1, "connectionID", connections.getString(i, "connectionID"));
            data.setInt(data.getRowCount()-1, "from", connections.getInt(i, "from"));
            data.setInt(data.getRowCount()-1, "to", connections.getInt(i, "to"));
            
            
            GregorianCalendar fromTime = new GregorianCalendar();
            GregorianCalendar toTime = new GregorianCalendar();
            
            try {
                
                Date readTime = dateTimeFormat.parse(
                        dateFormat.format(tempD.getTime())
                        + " " +
                        connections.getString(i, "fromTime"));
                
                fromTime.setTime(readTime);
                data.setDate(data.getRowCount()-1, "fromTimeReal", fromTime.getTime());
                
                readTime = dateTimeFormat.parse(
                        dateFormat.format(tempD.getTime())
                        + " " +
                        connections.getString(i, "toTime"));
                
                toTime.setTime(readTime);
                data.setDate(data.getRowCount()-1, "toTimeReal", toTime.getTime());
                
                long addMillis = 0;
                
                if (i > 0) {
                    
                    if (connections.getString(i-1, "connectionID").equals(connections.getString(i, "connectionID"))) {
                        
                        readTime = dateTimeFormat.parse(
                                dateFormat.format(tempD.getTime())
                                + " " +
                                connections.getString(i-1, "fromTime"));
                        addMillis = fromTime.getTimeInMillis() - readTime.getTime();
                    }
                }
                
                toTime.setTimeInMillis(tempD.getTimeInMillis() + toTime.getTimeInMillis() - fromTime.getTimeInMillis() + addMillis);
                fromTime.setTimeInMillis(tempD.getTimeInMillis() + addMillis);
                
            } catch (ParseException ex) {
                System.err.println("Can't read \"fromTime\"- or \"toTime\"-column form connections on line "
                        + i+2); // i beginns with 0, first line of file are header
            }
            
            data.setDate(data.getRowCount()-1, "toTime", toTime.getTime());
            data.setDate(data.getRowCount()-1, "fromTime", fromTime.getTime());
            
            
            
            // add the stopOvers if available
            if (connections.canGet("stopOvers", String.class)) {
                
                // render stopOvers
                
                
                String stopOversString = connections.getString(i, "stopOvers");
                if(stopOversString == null) stopOversString = "";
                if (stopOversString.endsWith("~")) stopOversString = stopOversString.substring(0, stopOversString.length()-1);
                String stopOvers[] = stopOversString.split("~");
                
                long diffMillis = data.getDate(data.getRowCount()-1, "toTimeReal").getTime() - toTime.getTimeInMillis();
                
                for (int j = 0; j < stopOvers.length; j++) {
                    
                    try {
                        
                        Date stopOverTime = dateTimeFormat.parse(
                                dateFormat.format(toTime.getTimeInMillis())
                                + " " +
                                stopOvers[j]);
                        
                        stopOvers[j] = timeFormat.format(stopOverTime.getTime() - diffMillis);
                    } catch (ParseException ex) {
                        System.err.println("Can't recalculate \"stopOvers\"");
                    }
                    
                }
                
                stopOversString = "";
                
                for (int k = 0; k < stopOvers.length; k++) stopOversString += stopOvers[k] + "~";
                
                data.setString(data.getRowCount()-1, "stopOvers", stopOversString);
                
            }
            
            
            
            
            
            
            data.setString(data.getRowCount()-1, "trainID", connections.getString(i, "trainID"));
            
            
        }
        
        
        // debug
        if (false) {
            
            for (int i = 0; i < data.getRowCount(); i++)
                System.out.println( "Nr.: " + (i+1) +
                        " fromTime: " + data.getDate(i, "fromTime").toString() +
                        " toTime: " + data.getDate(i, "toTime").toString() );
        }
        
        return data;
    }
    
    
    
    /**
     * Return true if a connection goes on a specific date
     * @param connectID ID of the connection
     * @param onDate The specific Date
     * @param dates Table of Dates
     * @return True if the Connection goes on the specific date
     */
    public static boolean trainOnDate(int connectID, GregorianCalendar onDate, Table dates) {
        
        boolean trainOnDate = false;
        onDate.setFirstDayOfWeek(GregorianCalendar.MONDAY);
        
        // i = 1, because the first line ist the date of planevalue
        for (int i = 1; i < dates.getRowCount(); i++) {
            
            if (dates.getInt(i, "connectID") == connectID) {
                
                if (dates.getString(i, "weakly") != "0000000") {
                    
                    char[] weakly = dates.getString(i, "weakly").toCharArray();
                    
                    try{
                        
                        char[] yes = "1".toCharArray();
                        if (weakly[onDate.get(GregorianCalendar.DAY_OF_WEEK)-1] == yes[0]) trainOnDate = true;
                    } catch (Exception e) {}
                } else {
                    
                    GregorianCalendar date_temp = new GregorianCalendar();
                    
                    date_temp.setTime(dates.getDate(i, "specialDate"));
                    if (onDate.compareTo(date_temp) == 0) return true;
                    
                    date_temp.setTime(dates.getDate(i, "notOnDate"));
                    if (onDate.compareTo(date_temp) == 0) return false;
                }
            }
        }
        
        return trainOnDate;
    }
    
    /**
     * Return a string with traininfos abaut the plan
     * @return String with traininfos
     * @param connectID ID of the connection
     * @param dates Table of Dates
     */
    public static String trainOnDateInfo(String connectID, Table dates) {
        
        String trainOnDateInfo = "";
        String trainOnSpecialDate = "";
        String trainNotOnSpecialDate= "";
        
        // i = 1, because the first line ist the date of planevalue
        for (int i = 1; i < dates.getRowCount(); i++) {
            
            if (dates.getString(i, "connectID").equals(connectID)) {
                
                if (!dates.getString(i, "weakly").equals("0") &&
                        !dates.getString(i, "weakly").equals("0000000") ) {
                    
                    String weaklyString = dates.getString(i, "weakly");
                    
                    if (weaklyString.length() != 6) {
                        
                        while (weaklyString.length() < 7)
                            weaklyString = "0" + weaklyString;
                    }
                    
                    if (weaklyString.equals("1111111"))  trainOnDateInfo = "tglich";
                    else if (weaklyString.equals("1111100"))  trainOnDateInfo = "Montag bis Freitag";
                    else if (weaklyString.equals("1111110"))  trainOnDateInfo = "Montag bis Samstag";
                    else {
                        
                        char[] weakly = weaklyString.toCharArray();
                        
                        if (weakly[0] == '1') trainOnDateInfo += "Montag, ";
                        if (weakly[1] == '1') trainOnDateInfo += "Dienstag, ";
                        if (weakly[2] == '1') trainOnDateInfo += "Mittwoch, ";
                        if (weakly[3] == '1') trainOnDateInfo += "Donnerstag, ";
                        if (weakly[4] == '1') trainOnDateInfo += "Freitag, ";
                        if (weakly[5] == '1') trainOnDateInfo += "Samstag, ";
                        if (weakly[6] == '1') trainOnDateInfo += "Sonntag, ";
                        
                        if (trainOnDateInfo.length() > 2) trainOnDateInfo = trainOnDateInfo.substring(0, trainOnDateInfo.length()-2);
                        
                        // replace last ", " with " und "
                        if (trainOnDateInfo.indexOf(", ") >= 0)
                            trainOnDateInfo = trainOnDateInfo.substring(0, trainOnDateInfo.lastIndexOf(", ")) +
                                    " und " +
                                    trainOnDateInfo.substring(trainOnDateInfo.lastIndexOf(", ") + ", ".length());
                        else trainOnDateInfo = " nur " + trainOnDateInfo;
                    }
                    
                } else {
                    
                    SimpleDateFormat dateFormat = new SimpleDateFormat( "dd.MM.yy" );
                    SimpleDateFormat shortDateFormat = new SimpleDateFormat( "dd.MM." );
                    
                    try{
                        //System.out.println(dates.getString(i, "specialDate"));
                        if (!dates.getString(i, "specialDate").equals("30.11.02") &&
                                !dates.getString(i, "specialDate").equals("00.00.0000")  )
                            trainOnSpecialDate += shortDateFormat.format(dateFormat.parse(dates.getString(i, "specialDate"))) + ", ";
                        
                        if (!dates.getString(i, "notOnDate").equals("30.11.02") &&
                                !dates.getString(i, "notOnDate").equals("00.00.0000") )
                            trainNotOnSpecialDate += shortDateFormat.format(dateFormat.parse(dates.getString(i, "notOnDate"))) + ", ";
                        
                    } catch (Exception e) { System.err.println("Can't read times form dates"); }
                }
            }
        }
        
        if (trainOnSpecialDate.length() > 2) trainOnSpecialDate = trainOnSpecialDate.substring(0, trainOnSpecialDate.length()-2);
        if (trainNotOnSpecialDate.length() > 2) trainNotOnSpecialDate = trainNotOnSpecialDate.substring(0, trainNotOnSpecialDate.length()-2);
        
        trainOnDateInfo = "fhrt " + trainOnDateInfo + "<br>";
        if (trainOnSpecialDate.length() > 0) trainOnDateInfo += " und am " + trainOnSpecialDate + "<br>";
        if (trainNotOnSpecialDate.length() > 0) trainOnDateInfo += " und NICHT am " + trainNotOnSpecialDate;
        
        return trainOnDateInfo;
    }
    
    
    
    /**
     * Adds internal Connection IDs, ids are 0, 10, 20, 30...
     * @return Table of data with internalConnectionIDs
     * @param connections Table of connections
     */
    public static Table addInternalConnectionIDs(Table connections) {
        
        Table connections_new = new Table();
        
        // copy columns
        for (int i = 0; i < connections.getColumnCount(); i++)
            connections_new.addColumn(connections.getColumnName(i), connections.getColumnType(i));
        
        // copy rows
        for (int i = 0; i < connections.getRowCount(); i++)
            connections_new.addTuple(connections.getTuple(i));
        
        connections_new.addColumn("internalConnectionID", int.class );
        
        String lastConnectionID = null;
        int connectionIdCounter = 0;
        
        for (int i = 0; i < connections_new.getRowCount(); i++) {
            
            if (i == 0 || !connections_new.getString(i, "connectionID").equals(lastConnectionID)) connectionIdCounter+=10;
            
            lastConnectionID = connections_new.getString(i, "connectionID");
            connections_new.setInt(i, "internalConnectionID", connectionIdCounter);
        }
        
        return connections_new;
    }
    
    
    /**
     * Adds new columns with trainnames and infos on the data-table
     * @param data Table of connections
     * @param trains Table of traininfos
     * @return Table of data with trainnames and infos added
     */
    public static Table addTrainInfo(Table data, Table trains) {
        
        data.addColumn("trainName", String.class, " ");
        data.addColumn("trainInfo", String.class, " ");
        
        for (int i = 0; i < data.getRowCount(); i++) {
            
            for (int j = 0; j < trains.getRowCount(); j++) {
                
                if (data.getString(i, "trainID").equals(trains.getString(j, "trainID"))) {
                    
                    data.setString(i, "trainName", trains.getString(j, "trainName"));
                    
                    String trainInfo = trains.getString(j, "trainInfo").replaceAll("~", "\n");
                    trainInfo = trainInfo.replaceAll("\n ", "\n");
                    if (trainInfo.endsWith("\n")) trainInfo = trainInfo.substring(0, trainInfo.length()-1);
                    
                    data.setString(i, "trainInfo", trainInfo);
                }
            }
        }
        
        return data;
    }
    
    
    
    /**
     * Adds a new column "tipText" with traininfos in HTML specially for tooltips
     * @return Table of data with trainnames and infos added
     * @param dates Table of dates
     * @param relativ If relativ, timeinfos ware taken from "fromTimeReal" and "toTimeReal"
     * @param data Table of connections with tiptextrow
     */
    public static Table addTipText(Table data, Table dates, boolean relativ) {
        
        data.addColumn("tipText", String.class, "no Info");
        
        for (int i = 0; i < data.getRowCount(); i++) {
            
            SimpleDateFormat timeFormat = new SimpleDateFormat( "HH:mm" );
            String tipText =
                    "<HTML>" +
                    "<span style=\"font-family:arial\">" +
                    "<b>" +
                    "<font size=\"+1\">" +
                    data.getString(i, "trainName") +
                    "</font>" +
                    "</b>" +
                    "&nbsp;&nbsp;&nbsp;&nbsp;";
            
            if (data.canGetDate("fromTimeReal") && data.canGetDate("toTimeReal")) {
                
                tipText +=
                        timeFormat.format(data.getDate(i, "fromTimeReal")) +
                        " - " +
                        timeFormat.format(data.getDate(i, "toTimeReal")) +
                        " &nbsp;&nbsp;&nbsp;&nbsp;(" +
                        //timeFormat.format(data.getDate(i, "toTime")) +
                        timeFormat.format(data.getDate(i, "toTime").getTime() - data.getDate(i, "fromTime").getTime() - 1000*60*60) +
                        ")";
            } else {
                
                tipText +=
                        timeFormat.format(data.getDate(i, "fromTime")) +
                        " - " +
                        timeFormat.format(data.getDate(i, "toTime")) +
                        " &nbsp;&nbsp;&nbsp;&nbsp;(" +
                        timeFormat.format(data.getDate(i, "toTime").getTime() - data.getDate(i, "fromTime").getTime() - 1000*60*60) +
                        ")";
            }
            
            tipText +=  "<br>" +
                    "<span style=\"text-align:center;\">" +
                    "<b>" +
                    data.getString(i, "fromName") +
                    "&nbsp;&nbsp; - &nbsp;&nbsp;" +
                    data.getString(i, "toName") +
                    "</b>" +
                    "</span><br>" +
                    "<hr>" +
                    "  - " +
                    data.getString(i, "trainInfo").replaceAll("\n", "<br>  - ") +
                    "</span>" +
                    "<hr>" +
                    trainOnDateInfo(data.getString(i, "connectionID"), dates) +
                    "</HTML>";
            
            data.setString(i, "tipText", tipText);
        }
        
        return data;
    }
    
    
    /**
     * Adds new columns with trainnames on the data-tabel
     * @return Table of data with stationnames added
     * @param stations Table of stationnames and infos
     * @param data Table of connections
     */
    public static Table addStationNames(Table data, Table stations) {
        
        data.addColumn("fromName", String.class, " ");
        data.addColumn("fromKm", int.class);
        data.addColumn("toName", String.class, " ");
        data.addColumn("toKm", int.class);
        
        
        for (int i = 0; i < data.getRowCount(); i++) {
            
            for (int j = 0; j < stations.getRowCount(); j++) {
                
                if (data.getString(i, "from").equals(stations.getString(j, "stationID"))) {
                    
                    data.setString(i, "fromName", stations.getString(j, "stationName"));
                    data.setInt(i, "fromKm", stations.getInt(j, "stationKM"));
                }
                
                if (data.getString(i, "to").equals(stations.getString(j, "stationID"))) {
                    
                    data.setString(i, "toName", stations.getString(j, "stationName"));
                    data.setInt(i, "toKm", stations.getInt(j, "stationKM"));
                }
            }
        }
        
        return data;
    }
    
    
    /**
     * Get the earliste connection (time/date) of a specific date
     * @param data Table of data/connections
     * @param day the specific day
     * @return date/time of eralist connection
     */
    public static GregorianCalendar getEarliestDate(Table data, GregorianCalendar day) {
        
        GregorianCalendar returnDate = new GregorianCalendar();
        GregorianCalendar tempD = new GregorianCalendar();
        
        for (int i = 0; i < data.getRowCount(); i++) {
            
            tempD.setTime(data.getDate(i, "fromTime"));
            tempD.set(tempD.get(tempD.YEAR), tempD.get(tempD.MONTH), tempD.get(tempD.DAY_OF_MONTH), 0, 0, 0);
            
            if ( day.get(tempD.YEAR) == tempD.get(tempD.YEAR) &&
                    day.get(tempD.MONTH) == tempD.get(tempD.MONTH) &&
                    day.get(tempD.DAY_OF_MONTH) == tempD.get(tempD.DAY_OF_MONTH) ) {
                
                if (i == 0) returnDate.setTime(data.getDate(i, "fromTime"));
                
                tempD.setTime(data.getDate(i, "fromTime"));
                
                if (tempD.compareTo(returnDate) < 0) returnDate.setTime(tempD.getTime());

            }
        }
        
        return returnDate;
    }
    
    
    
    /**
     * Get the lastest arival of connection (time/date) of a specific date
     * @param data Table of data/connections
     * @param day the specific day
     * @return date/time of arrival of the last connection
     */
    public static GregorianCalendar getLatestDate(Table data, GregorianCalendar day) {
        
        GregorianCalendar returnDate = new GregorianCalendar();
        GregorianCalendar tempD = new GregorianCalendar();
        
        for (int i = 0; i < data.getRowCount(); i++) {
            
            tempD.setTime(data.getDate(i, "toTime"));
            tempD.set(tempD.get(tempD.YEAR), tempD.get(tempD.MONTH), tempD.get(tempD.DAY_OF_MONTH), 0, 0, 0);
            
            if ( day.get(tempD.YEAR) == tempD.get(tempD.YEAR) &&
                    day.get(tempD.MONTH) == tempD.get(tempD.MONTH) &&
                    day.get(tempD.DAY_OF_MONTH) == tempD.get(tempD.DAY_OF_MONTH) ) {
                
                if (i == 0) returnDate.setTime(data.getDate(i, "toTime"));
                
                tempD.setTime(data.getDate(i, "toTime"));
                
                if (tempD.compareTo(returnDate) > 0) returnDate.setTime(tempD.getTime());
            }
        }
        
        return returnDate;
    }
    
    
    /**
     * Sets the calendar MinSelectableDate and MaxSelectableDate
     * @param calendar Calendar to set
     * @param dates Table of dates
     * @return Modified calendar
     */
    public static JCalendar setValidDates(JCalendar calendar, Table dates) {
        
        SimpleDateFormat dateFormat = new SimpleDateFormat( "dd.MM.yy" );
        
        GregorianCalendar fromDate = new GregorianCalendar();
        GregorianCalendar toDate = new GregorianCalendar();
        
        try{
            
            fromDate.setTime(dateFormat.parse(dates.getString(0, "specialDate")));
            toDate.setTime(dateFormat.parse(dates.getString(0, "notOnDate")));
        } catch (Exception e) {}
        
        calendar.setMinSelectableDate(fromDate.getTime());
        calendar.setMaxSelectableDate(toDate.getTime());
        
        return calendar;
    }
    
    
    /**
     * convert an object to a long
     */
    static public long object2Long(Object obj) {
        if (obj == null) return 0;
        if (obj instanceof Number)
            return ((Number)obj).longValue();
        else
            return Long.parseLong(obj.toString());
    }
    
    /**
     * Returns the int value that is closest to l.  That is, if l can fit into a
     * 32-bit unsigned number, returns (int)l.  Otherwise, returns either
     * Integer.MAX_VALUE or Integer.MIN_VALUE as appropriate.
     */
    public static int long2int(long l) {
        if (l>=Integer.MAX_VALUE)
            return Integer.MAX_VALUE;
        else if (l<=Integer.MIN_VALUE)
            return Integer.MIN_VALUE;
        else
            return (int)l;
    }
    
    
    
    /**
     *
     * @param ID
     * @param IDfield
     * @param field
     * @param t
     * @return
     */
    public static Object getFieldforID(String ID, String IDfield, String field, Table t) {
        Iterator it = t.tuples();
        
        while(it.hasNext()){
            Tuple act = (Tuple) it.next();
            
            if(act.getString(IDfield).equals(ID)) return act.get(field);
        }
        
        return null;
    }
    
    /**
     * find the last row in a table containing a given value
     * @param searchField the field of the table that is compared to the searchValue parameter.
     * @param searchValue the value that is searched
     */
    public static int searchLast(String searchField, String searchValue, Table t) {
        Iterator it = t.tuples();
        int row = -1;
        
        while(it.hasNext()){
            Tuple act = (Tuple) it.next();
            
            if(act.getString(searchField).equals(searchValue)) row = act.getRow();
            
        }
        
        return row;
    }
    
    
    /**
     * tries to extract the 4 necessary datatables from a given zip file
     * @return an empty string of no errors occurred. Otherwise an error message is returned.
     */
    public static String extractTables(String zipfileName, String tempDirName) {
        
        String fileSep = System.getProperty("file.separator");
        String errorMsg = "";
        
        try {
            
            File tempDir = new File(tempDirName);
            deleteDirectory(tempDir);
            tempDir.mkdirs();
            
            File zipFile = new File(zipfileName);
            ZipInputStream in = null;
            if (zipFile.exists()) in = new ZipInputStream(new FileInputStream(zipfileName));
            else in = new ZipInputStream(Functions.class.getClassLoader().getResourceAsStream(zipfileName));
           
            ZipEntry entry = in.getNextEntry();
            
            while (entry != null) {
                
                if (entry.getName().endsWith("connections.csv") ||
                        entry.getName().endsWith("dates.csv") ||
                        entry.getName().endsWith("stations.csv") ||
                        entry.getName().endsWith("trains.csv") ) {
                    
                    File tempZipFile = new File(entry.getName());

                    File outPutFile = new File(tempDir.getName() + fileSep + tempZipFile.getName());
                    if (outPutFile.exists()) outPutFile.delete();
                    
                    // Open the output file
                    OutputStream out = new FileOutputStream(outPutFile);
                    
                    // Transfer bytes from the ZIP file to the output file
                    byte[] buf = new byte[1024];
                    int len;
                    while ((len = in.read(buf)) > 0) {
                        out.write(buf, 0, len);
                    }
                    
                    // Close the outputfile
                    out.close();
                }
                
                entry = in.getNextEntry();
            }
            // Close inputfile
            in.close();
            
            if ( !new File(tempDir.getName() + fileSep + "connections.csv").exists() ) {
                
                errorMsg += "\"Connections.csv\" not found in zipfile~";
            }
            
            if ( !new File(tempDir.getName() + fileSep + "dates.csv").exists() ) {
                
                errorMsg += "\"dates.csv\" not found in zipfile~";
            }
            
            if ( !new File(tempDir.getName() + fileSep + "stations.csv").exists() ) {
                
                errorMsg += "\"stations.csv\" not found in zipfile~";
            }
            
            if ( !new File(tempDir.getName() + fileSep + "trains.csv").exists() ) {
                
                errorMsg += "\"trains.csv\" not found in zipfile~";
            }
            
        } catch (Exception ex) {ex.getStackTrace(); errorMsg += "Could not extract zipfile~"; }
        
        if (!errorMsg.equals("")) return errorMsg;
        else return "";
    }
    
    
    public static void deleteDirectory(File path) {
        
        if( path.exists() ) {
            File[] files = path.listFiles();
            for(int i=0; i<files.length; i++) {
                if(files[i].isDirectory()) {
                    deleteDirectory(files[i]);
                } else {
                    files[i].delete();
                }
            }
        }
        path.delete();
    }
    
    
    
    /**
     * tests the consitency of the 4 datatables
     */
    public static String validateTables(Table connections, Table stations, Table trains, Table dates) {
        
        String errorMsg = "";
        
        try {
            
            // Container for IDs
            Vector trainIDs = new Vector();
            Vector stationIDs = new Vector();

            // connections
            if (!connections.canGetInt("connectionID")) errorMsg += "\"connections.csv\" - Errors in column \"connectionID\" (Integer)~";
            if (!connections.canGetInt("trainID")) errorMsg += "\"connections.csv\" - Errors in column \"trainID\" (Integer)~";
            if (!connections.canGetInt("from")) errorMsg += "\"connections.csv\" - Errors in column \"from\" (Integer)~";
            if (!connections.canGetString("fromTime")) errorMsg += "\"connections.csv\" - Errors in column \"fromTime\" (String)~";
            if (!connections.canGetInt("to")) errorMsg += "\"connections.csv\" - Errors in column \"to\" (Integer)~";
            if (!connections.canGetString("toTime")) errorMsg += "\"connections.csv\" - Errors in column \"toTime\" (String)~";



            // stations
            if (!stations.canGetInt("stationID")) errorMsg += "\"stations.csv\" - Errors in column \"stationID\" (Integer)~";
            else for (int i = 0; i < stations.getRowCount(); i++) 
                if (!stationIDs.contains(stations.getInt(i, "stationID"))) 
                    stationIDs.add(stations.getInt(i, "stationID"));        
            if (!stations.canGetString("stationName")) errorMsg += "\"stations.csv\" - Errors in column \"stationName\" (String)~";
            //if (!stations.canGetString("stationInfo")) errorMsg += "\"stations.csv\" - Errors in column \"stationInfo\" (String)~";
            if (!stations.canGetInt("stationKM")) errorMsg += "\"stations.csv\" - Errors in column \"stationKM\" (Integer)~";

            // trains
            if (!trains.canGetInt("trainID")) errorMsg += "\"trains.csv\" - Errors in column \"trainID\" (Integer)~";
            else for (int i = 0; i < trains.getRowCount(); i++) 
                if (!trainIDs.contains(trains.getInt(i, "trainID"))) 
                    trainIDs.add(trains.getInt(i, "trainID"));        
            if (!trains.canGetString("trainName")) errorMsg += "\"trains.csv\" - Errors in column \"trainName\" (String)~";
            if (!trains.canGetString("trainInfo")) errorMsg += "\"trains.csv\" - Errors in column \"trainInfo\" (String)~";

            // dates
            if (!dates.canGetInt("connectID")) errorMsg += "\"dates.csv\" - Errors in column \"connectID\" (Integer)~";
            if (!dates.canGetString("weakly")) errorMsg += "\"dates.csv\" - Errors in column \"weakly\" (String)~";
            if (!dates.canGetString("specialDate")) errorMsg += "\"dates.csv\" - Errors in column \"specialDate\" (String)~";
            if (!dates.canGetString("notOnDate")) errorMsg += "\"dates.csv\" - Errors in column \"notOnDate\" (String)~";

            // connections
            for (int i = 0; i < connections.getRowCount(); i++) {

                if (!trainIDs.contains(connections.getInt(i, "trainID"))) 
                    errorMsg += "\"connections.csv\" - Error on line " + (i+1) + " in column \"trainID\", this train not exists in \"trains.csv\" (column trainID)~";

                if (!stationIDs.contains(connections.getInt(i, "from"))) 
                    errorMsg += "\"connections.csv\" - Error on line " + (i+1) + " in column \"from\", this station not exists in \"stations.csv\" (column stationID)~";

                if (!stationIDs.contains(connections.getInt(i, "to"))) 
                    errorMsg += "\"connections.csv\" - Error on line " + (i+1) + " in column \"to\", this station not exists in \"stations.csv\" (column stationID)~";            
            }

            // stations
            for (int i = 0; i < stations.getRowCount(); i++)            
                if (i > 0 && stations.getInt(i, "stationKM") <= stations.getInt(i-1, "stationKM") )
                    errorMsg += "\"stations.csv\" - Error on line " + (i+1) + " in column \"stationKM\", preceding km are greater~";
        }
        catch (Exception e) { errorMsg += "Validating tables aborted (unknown error)~"; }
        
        
        if (!errorMsg.equals("")) return "Errors while validating tables~" + errorMsg;
        else return "";
    }
    
    
    
    
    
}
