/*
 * LineRenderer.java
 *
 * Created on June 29, 2007, 9:32 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package zugvis;

import at.ac.tuwien.cs.timevis.FisheyeTimeScale;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RectangularShape;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.awt.Rectangle;
import java.awt.Point;

import prefuse.data.Tuple;
import prefuse.Constants;
import prefuse.render.AbstractShapeRenderer;
import prefuse.util.ColorLib;
import prefuse.util.GraphicsLib;
import prefuse.visual.VisualItem;
import prefuse.Visualization;
import prefuse.visual.VisualTable;

/**
     * This class is responsible for the rendering of visualitems that belong to the connections Table. For each item a line 
     * is drawn and at the start and endpoint of the line a small circle is placed indicating that the train stops at the
     * current station. The X-axis represents the time and the Y-axis represents the trainstations.
     */
public class LineRenderer extends AbstractShapeRenderer {
    
        private static final Color trainSelectedColor = new Color(255, 0, 0);
        
        private FisheyeTimeScale timeScale = null;
        private boolean drawStations = true;
        private Visualization MyVis;
        private String group;
        
        public LineRenderer(String m_group, Visualization vis){
            MyVis = vis;
            group = m_group;
        }
        
        public void setTimeScale(FisheyeTimeScale ts){
            timeScale = ts;
        }
        
        /**
         * this sets the drawstations flag. If true stations of a connection will be drawn as small white lines.
         */
        public void drawStations(boolean draw){
            drawStations = draw;
        }
        /**
         * returns the Boundingbox of a given VisualItem
         */
       public Shape getShape(VisualItem item){
           Rectangle rect = getRawShape(item).getBounds();
           rect.height += 5;
           if(rect.width < 10){
               rect.width = 20;
               rect.x -= 8;
           }
           return rect;
           
        }
        
        /**
         * return the line in the x,y coordinate system of a given visualitem
         */
        public Shape getTrainLine(VisualItem item){
            return getRawShape(item);
        }
        
        /**
         * return the line in the x,y coordinate system of a given visualitem
         */
        protected Shape getRawShape(VisualItem item) {
            Line2D.Double line = new Line2D.Double();
            Tuple Zug = item.getSourceTuple();
            
            double x1 = item.getX();
            double y1 = item.getY();
            double x2 = timeScale.getPixelForDate(item.getDate("toTime").getTime());
            double y2 = item.getDouble("toY");
            
            
            if ( Double.isNaN(x1) || Double.isInfinite(x1) ) x1 = 0;
            if ( Double.isNaN(y1) || Double.isInfinite(y1) ) y1 = 0;
            if ( Double.isNaN(x2) || Double.isInfinite(x2) ) x2 = 0;
            if ( Double.isNaN(y2) || Double.isInfinite(y2) ) y2 = 0;
            
            line.setLine(x1, y1, x2, y2);
            
            return line;
        }
        
        /**
         * This method draws a given visualitem into a Graphics2D context. The getRawShape-method is used to determine the position 
         * of the item. Every item is supposed to be a line and at its start and endpoint a small circle is drawn.
         * The color for the line comes from the StrokeColor field of the visualitem and the color for the circles comes from
         * the FillColor field.
         */
        protected void drawShape(java.awt.Graphics2D g, VisualItem item, Shape shape){
            Color color = new Color(item.getStrokeColor());
            if(item.isHighlighted()) color = trainSelectedColor;
            g.setPaint(color);
            shape = getRawShape(item);
            g.draw(shape);
            
            VisualTable t = (VisualTable) MyVis.getGroup(group);
            
            if(t.getRowCount() != (item.getRow() + 1) ){
                if(t.getInt(item.getRow() + 1, "ID") == item.getInt("ID")){
                    Line2D.Double line = (Line2D.Double) getRawShape(t.getItem(item.getRow() + 1));
                    Line2D.Double actItem = (Line2D.Double) shape;
                    line.x2 = actItem.x2;
                    line.y2 = actItem.y2;
                    
                    g.draw(line);
                    Point p = new Point((int) (line.x1 + (line.x2 - line.x1) / 2), (int) line.y1);
                    item.set("toPoint", p);
                    t.getItem(item.getRow() + 1).set("fromPoint", p);
                    
                    double xdiff = (line.x1 - line.x2) / 2;
                    if(item.getStrokeColor() == item.getFillColor() || item.isHighlighted() || drawStations) g.fill(new java.awt.geom.Ellipse2D.Double(line.x1 - xdiff - 4, line.y2 - 4, 8, 8));
                }else{
                    if(item.getStrokeColor() == item.getFillColor() || item.isHighlighted() || drawStations) g.fill(new java.awt.geom.Ellipse2D.Double(((Line2D)shape).getX2() - 4, ((Line2D)shape).getY2() - 4, 8, 8));
                }
                if(item.getRow() != 0){
                    if(t.getInt(item.getRow() - 1, "ID") != item.getInt("ID")){
                        if(item.getStrokeColor() == item.getFillColor() || item.isHighlighted() || drawStations) g.fill(new java.awt.geom.Ellipse2D.Double(((Line2D)shape).getX1() - 4, ((Line2D)shape).getY1() - 4, 8, 8));
                    }
                }
            }else{
                if(item.getStrokeColor() == item.getFillColor() || item.isHighlighted() || drawStations) g.fill(new java.awt.geom.Ellipse2D.Double(((Line2D)shape).getX2() - 4, ((Line2D)shape).getY2() - 4, 8, 8));
            }
            
            
        }
    };