/*
 * LineVis.java
 *
*/

package zugvis;

import at.ac.tuwien.cs.timevis.BasicTimeScale;
import at.ac.tuwien.cs.timevis.FisheyeTimeScale;
import at.ac.tuwien.cs.timevis.actions.PanAction;
import at.ac.tuwien.cs.timevis.actions.ZoomAction;
import at.ac.tuwien.cs.timevis.prefuse.TimeLayout;
import at.ac.tuwien.cs.timevis.ui.FisheyeSlider;
import at.ac.tuwien.cs.timevis.ui.MouseTracker;
import at.ac.tuwien.cs.timevis.ui.StatusBar;
import at.ac.tuwien.cs.timevis.ui.TimeScaleHeader;
import at.ac.tuwien.cs.timevis.ui.TimeScalePainter;


import prefuse.Display;
import prefuse.data.Table;
import prefuse.data.Schema;
import prefuse.data.expression.parser.JavaCharStream;
import prefuse.data.tuple.TableTuple;
import prefuse.data.Tuple;
import prefuse.data.tuple.TupleSet;
import prefuse.data.query.RangeQueryBinding;
import prefuse.data.query.NumberRangeModel;
import prefuse.Visualization;
import prefuse.render.AbstractShapeRenderer;
import prefuse.render.DefaultRendererFactory;
import prefuse.render.RendererFactory;
import prefuse.render.Renderer;
import prefuse.render.ShapeRenderer;
import prefuse.render.AxisRenderer;
import prefuse.render.LabelRenderer;
import prefuse.visual.VisualItem;
import prefuse.visual.VisualTable;
import prefuse.visual.sort.ItemSorter;
import prefuse.util.PrefuseLib;
import prefuse.util.ColorLib;
import prefuse.util.ui.ValuedRangeModel;
import prefuse.util.ui.JFastLabel;
import prefuse.action.*;
import prefuse.action.layout.Layout;
import prefuse.action.layout.graph.ForceDirectedLayout;
import prefuse.action.layout.AxisLayout;
import prefuse.action.layout.AxisLabelLayout;
import prefuse.activity.Activity;
import prefuse.action.assignment.ColorAction;
import prefuse.action.assignment.DataColorAction;
import prefuse.action.assignment.ShapeAction;
import prefuse.visual.expression.VisiblePredicate;
import prefuse.Constants;
import prefuse.controls.Control;
import prefuse.controls.ControlAdapter;
import prefuse.controls.PanControl;
import prefuse.controls.ZoomControl;
import prefuse.util.ui.UILib;
import prefuse.util.FontLib;
import prefuse.util.display.PaintListener;
import prefuse.util.ui.JRangeSlider;
import java.util.Vector;

import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JTabbedPane;
import javax.swing.JComponent;
import javax.swing.JToolBar;
import javax.swing.BorderFactory;
import javax.swing.SwingConstants;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.event.ChangeEvent;
import javax.swing.JToolTip;
import javax.swing.event.ChangeListener;
import javax.swing.SwingUtilities;
import java.awt.event.ComponentEvent;
import java.awt.geom.Line2D;
import java.awt.BorderLayout;
import java.awt.Shape;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Ellipse2D;
import java.awt.Color;
import java.awt.event.MouseEvent;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.awt.event.ItemListener;
import java.awt.event.ComponentAdapter;
import java.awt.FontMetrics;
import java.text.ChoiceFormat;
import java.text.NumberFormat;
import java.text.FieldPosition;
import java.text.ParsePosition;
import java.lang.StringBuffer;
import java.util.Iterator;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;


/**
 * Class that encapsulates the EJ. Marey's Train Schedule
 * @author Georg Prohaska
 */
public class LineVis extends JPanel{
    
    /**
     * color of the highlights that are used when the mouse moves over an item
     */
    private static final Color highlightColor = new Color(255, 0, 0);
    private static final int highlightColorInt = ColorLib.rgb(250, 100, 0);
    private static final int trainColor = ColorLib.rgb(50,50,125);
    private static final Color stationHighlightColor = new Color(0, 255, 0);

    /**
     * Bounding Box for the stations. This is only used for the mouse-ever event
     */
    private static final Rectangle stationBounds = new Rectangle(10, 10);
    
    /**
     * size of the highlight
     */
    private static final int highlightRadius = 15;
    
    /**
     * X position of the border of the y axis. A line is drawn at that position and all trains at the left of it wont be drawn
     */
    private static final int yAxisBorder = 150;
    
    
    private static Visualization MyVis; //the Visualization-Object for this Class
    private static LineVisDisplay d; //the Display-object for this class 
    private static LineRenderer trainRenderer; //renderer for the connections Table
    private static DotRenderer dotrenderer = new DotRenderer(); //renderer for the highlight
    private static Display display; //the Display-object for this visualization
    private static NumberRangeModel TimeRange; //rangemodel of the X-axis
    private static NumberRangeModel LocationRange; //rangemodel of the Y-axis
    private static ChoiceFormat stationsFormat; //maps stationIDs to stationnames
    private static JTabbedPane tabPane; //pointer to the tabbedPane of the application

    private static boolean UPorDOWN = true; //flag for the direction. Determines whether station 1 is at the bottom or at the top of the display

    //the 4 tables containing all necessary data - "database" of the visualization
    private static Table connectionsTable = null;
    private static Table stationsTable = null;
    private static Table trainsTable = null;
    private static Table datesTable = null;
    
    private static Vector selectedTrains = null; //pointer to the vector that holds the trainIDs of the currently selected trains.
    
    /*
     * creates the RangeSlider
     */
    static JRangeSlider makeJRangeSlider_Y() {
        
        final JRangeSlider jRangeSlider_Y = new JRangeSlider(
                d.toLay.getRangeModel().getMinimum(),
                d.toLay.getRangeModel().getMaximum(),
                d.toLay.getRangeModel().getMinimum(),
                d.toLay.getRangeModel().getMaximum()-d.toLay.getRangeModel().getMinimum(),
                JRangeSlider.VERTICAL, 1);
        
        jRangeSlider_Y.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent evt) {
                if(!UPorDOWN){
                    d.toLay.getRangeModel().setValue(jRangeSlider_Y.getLowValue());
                    d.toLay.getRangeModel().setExtent(jRangeSlider_Y.getHighValue()-jRangeSlider_Y.getLowValue());
                    d.fromLay.getRangeModel().setValue(jRangeSlider_Y.getLowValue());
                    d.fromLay.getRangeModel().setExtent(jRangeSlider_Y.getHighValue()-jRangeSlider_Y.getLowValue());
                }else {
                    d.toLay.getRangeModel().setValue(jRangeSlider_Y.getMaximum() - jRangeSlider_Y.getHighValue());
                    d.toLay.getRangeModel().setExtent(jRangeSlider_Y.getHighValue()-jRangeSlider_Y.getLowValue());
                    d.fromLay.getRangeModel().setValue(jRangeSlider_Y.getMaximum() - jRangeSlider_Y.getHighValue());
                    d.fromLay.getRangeModel().setExtent(jRangeSlider_Y.getHighValue()-jRangeSlider_Y.getLowValue());
                }
                    
                MyVis.run("layout");
            }
        });
        
        return jRangeSlider_Y;
    }
  
    /**
     * An action that can flip the Y values of all trains. It is used to set the direction of the route.
     */
    private static class FlipDisplay extends GroupAction{
        
        /**
         * creates a new instance of FlipDisplay
         * @param group name of the group which holds the data.
         */
        public FlipDisplay(String group){
            super(group);
        }
        
        public void run(double frac){
            if(!UPorDOWN) return;
            Iterator it = ((VisualTable) MyVis.getGroup(m_group)).tuples();
            
            while(it.hasNext()){
                VisualItem act = (VisualItem) it.next();
                act.setY(d.getHeight() - act.getY());
                act.setDouble("toY", d.getHeight() - act.getDouble("toY"));
            }
        }
    }
    
    /**
     * An action that is used to determine which directions are visible
     */
    private static class Direction extends GroupAction{
        /**
         * both directions are visible
         */
        public static final int UPDOWN = 0;
        /**
         * only one direction is visible
         */
        public static final int UP = 1;
        /**
         * only one direction is visible
         */
        public static final int DOWN  = 2;
        
        /**
         * current state
         */
        private int Direction = 0;
        
        /**
         * creates a new instance of Direction.
         * @param group name of the group which holds the data.
         * @param dir initial value for the directions that are visible
         */
        public Direction(String group, int dir){
            super(group);
            Direction = dir;
        }
        
        /**
         * sets the visible directions
         */
        public void setDirection(int d){
            Direction = d;
        }
        
        public void run(double frac){
            
            Iterator it = MyVis.visibleItems();
            
            while(it.hasNext()){
                VisualItem act = (VisualItem) it.next();
                if(!act.isInGroup(m_group)) continue;
                int from = act.getInt("from");
                int to = act.getInt("to");
                
                switch(Direction){
                    case UP:
                        act.setVisible(from < to);
                        break;
                        
                    case DOWN:
                        act.setVisible(from > to);
                        break;
                        
                    default:
                        act.setVisible(true);
                }
            }
        }
    }
           
    
    /**
     * A layout for the labeling of the y-axis. Draws one label and one line for each station. The y coordinate of the labels depend on the "stationKM" column of the stations-table.
     */
    private static class stationAxisLabelLayout extends Layout {
        /**
         * the table that holds all stations.
         */
        private Table stationT;
        
        /**
         * creates a new instance of stationAxisLabelLayout.
         * @param group name of the group which holds the data.
         * @param stations the stations table that holds information about all stations.
         */
        public stationAxisLabelLayout(String group, Table stations){
            super(group);
            stationT = stations;
        }
        
        public void run(double frac){
            Rectangle2D b = getLayoutBounds();
            Iterator stations = stationT.tuples();
            
            ValuedRangeModel range = d.toLay.getRangeModel();
            Integer HighValue = (Integer) range.getHighValue();
            Integer LowValue = (Integer) range.getLowValue();
            Integer span = HighValue - LowValue;
            int DisplayHeight = d.getHeight();
            
            VisualTable Axis = (VisualTable)MyVis.getGroup(m_group);
            Axis.clear();
            
            while( stations.hasNext()){
                VisualItem actLabel = Axis.addItem();
                VisualItem actLine = Axis.addItem();
                TableTuple actStation = (TableTuple) stations.next();
                double Y = 0;
                double actValue = actStation.getDouble("stationKM");
                
                if(actStation.getDouble("stationKM") > HighValue || actStation.getDouble("stationKM") < LowValue) continue;
                
                actLabel.setStartVisible(true);
                actLabel.setEndVisible(true);
                actLabel.set("_label", actStation.getString("stationName"));
                actLabel.set("_value", actValue);
                
                actLabel.set(VisualItem.X, b.getMinX());
                Y = (HighValue - actValue) * (b.getMaxY() / span);
                if(UPorDOWN) Y = DisplayHeight - Y;
                
                FontMetrics fm = Renderer.DEFAULT_GRAPHICS.getFontMetrics(actLabel.getFont());
                
                if(actStation.getRow() != stationT.getRowCount()-1) Y = Y - fm.getHeight() / 2;
                if(actStation.getRow() == 0) Y = Y - fm.getHeight() / 2;

                if(UPorDOWN && actStation.getRow() == stationT.getRowCount()-1) Y = Y - fm.getHeight();
                if(UPorDOWN && actStation.getRow() == 0) Y = Y + fm.getHeight();
            
                
                actLabel.set(VisualItem.Y, Y);
                
                actLine.setStartVisible(true);
                actLine.setEndVisible(true);
                actLine.set("_value", actValue);
                
               
                if(actStation.getRow() != stationT.getRowCount()-1){
                    actLine.set(VisualItem.X, b.getMinX() + fm.stringWidth(actStation.getString("stationName")));
                } else actLine.set(VisualItem.X, b.getMinX());
               
                
                actLine.set(VisualItem.X2, b.getMaxX());

                Y = (HighValue - actValue) * (b.getMaxY() / span);
                if(UPorDOWN) Y = DisplayHeight - Y;
                actLine.set(VisualItem.Y, Y);
                
                Y = (HighValue - actValue) * (b.getMaxY() / span);
                if(UPorDOWN) Y = DisplayHeight - Y;
                actLine.set(VisualItem.Y2, Y);

            }
        }
        
    }
    
    /**
     * A simple action that copys the Y value of all visualitems to the "toY" colum of the each item.
     */
    private static class YtoEndY extends GroupAction {
        
        /**
         * creates a new instance of YtoEnd
         * @param  name of the group which holds the data.
         */
        public YtoEndY(String group){
            super(group);
        }
        
        public void run(double frac){
            
            VisualTable connections = (VisualTable)MyVis.getGroup(m_group);

            Iterator items = connections.tuples();
            while( items.hasNext()){
                VisualItem item = (VisualItem)items.next();
                item.setDouble("toY", item.getY());
            }
        }
    }
    
    /**
     * A simple action that sets the visible tag of all items that lie on the left side of the yAxisBorder to false.
     */
    private static class BorderFilter extends GroupAction {
        
        /**
         * creates a new instance of BorderFilter
         * @param  name of the group which holds the data.
         */
        public BorderFilter(String group){
            super(group);
        }
        
        public void run(double frac){
            
            Iterator items = ((VisualTable)MyVis.getGroup("data")).tuples();
            
            while( items.hasNext()){
                VisualItem item = (VisualItem)items.next();
                if( item.getX() < yAxisBorder || item.getX() > MyVis.getDisplay(0).getWidth())
                    item.setVisible(false);
                else
                    item.setVisible(true);
            }
        }
    }
    
    /**
     * An action that manages the process of train-selection.
     */
    private static class SelectTrains extends GroupAction{
        
        public SelectTrains(String group){
            super(group);
        }
        
        public void run(double frac){
            Iterator items = ((VisualTable)MyVis.getGroup("data")).tuples();
            
            while( items.hasNext()) ((VisualItem)items.next()).setHighlighted(false);
                
            
            Iterator it = selectedTrains.iterator();
            
            while(it.hasNext()) selectTrain((String) it.next(), true);
        }
    }
    

    
    
    /**
     * This class is just a "dirty-trick" to draw dots into the visualization. It draws a dot for every visualitem at the position
     * specified in the X2,Y2 fields of the item.
     */
    public static class DotRenderer extends AbstractShapeRenderer {
        
        /**
         * returns an ellipse at the position specified in the X2,Y2 field of the given visualitem
         */
        protected Shape getRawShape(VisualItem item){
            return new Ellipse2D.Double(item.getDouble(VisualItem.X2) - highlightRadius / 2, item.getDouble(VisualItem.Y2) - highlightRadius / 2, highlightRadius, highlightRadius);
        }
        
        /**
         * Draws a circle at the specified coordinates of the visualitem. The color of the circle is determined by the highlightColor constant
         * and the size of the circle is determined by the highlightRadius constant of the LineVis class.
         */
        protected void drawShape(java.awt.Graphics2D g, VisualItem item, Shape shape){
            g.setPaint(highlightColor);
            g.fill(getRawShape(item));
        }
    }
    
    /**
     * Sets the highlighted flag of a train.
     * @param ID the trainID of the train
     * @param highlight the highlight flag of the train is set to this value
     */
    private static void selectTrain(String ID, boolean highlight){
        VisualTable alltrains = (VisualTable) MyVis.getGroup("data");
        
        Iterator items = alltrains.tuples();
        
        while(items.hasNext()){
            VisualItem act = (VisualItem) items.next();
            
            if(act.getString("trainID").equals(ID)){
                act.setHighlighted(highlight);
            }
        }
    }
    
    
    /**
     * set the FillColor and StrokeColor field of the given train
     * @param ID the ID of the train
     * @param Color Color that FillColor and StrokeColor are set to
     */
    private void highlightTrain(double ID, int Color) {
        VisualTable alltrains = (VisualTable) MyVis.getGroup("data");
        
        Iterator items = alltrains.tuples();
        
        while(items.hasNext()){
            VisualItem act = (VisualItem) items.next();
            
            if(act.getDouble("ID") == ID){
                act.setStrokeColor(Color);
                act.setFillColor(Color);
            }
        }
    }
    
    /**
     * sets the color of the given train back to normal
     * @param ID ID of the train
     */
    private void dehighlightTrain(double ID) {
        VisualTable alltrains = (VisualTable) MyVis.getGroup("data");
        
        Iterator items = alltrains.tuples();
        
        while(items.hasNext()){
            VisualItem act = (VisualItem) items.next();
            
            if(act.getDouble("ID") == ID){
                act.setStrokeColor(trainColor);
                act.setFillColor(0);
            }
        }
    }
    
    /**
     * changes the values "from" and "to" columns of a given table to the the according "stationKM" values in a given stations-table
     * @param t "from" and "to" values of this table are changed
     * @param stations "from" and "to" values are changed to the values of the "stationKM"-column of this table
     */
    private void changeStationIDtoStationKM(Table t, Table stations){
        
        Iterator it = t.tuples();
        
        while(it.hasNext()){
            TableTuple act = (TableTuple) it.next();
            act.setInt("from", stations.getInt(searchTable(Integer.toString(act.getInt("from")), "stationID", stations), "stationKM"));
            act.setInt("to", stations.getInt(searchTable(Integer.toString(act.getInt("to")), "stationID", stations), "stationKM"));
        }
    }
    
    /**
     * searches a table for a specific tuple
     * @param ID the ID of the tuple that shall be found
     * @param IDColumn the name of the colum that holds the IDs of each tuple
     * @param t the table to be searched
     * @return returns the row of the tuple with the given ID
     */
    public int searchTable(String ID, String IDColumn, Table t){
        Iterator it = t.tuples();
        
        while(it.hasNext()){
            Tuple act = (Tuple)it.next();
            if(act.getString(IDColumn).equals(ID))
                return act.getRow();
        }
        return 0;
    }
    
    /**
     * sets the position of the first highlight
     * @param X X coordinate in pixels of the new position
     * @param Y Y coordinate in pixels of the new position
     */
    private void setHighlight1(double X, double Y){
        VisualTable hi = (VisualTable) MyVis.getGroup("highlight");
        hi.setDouble(0, VisualItem.X2, X);
        hi.setDouble(0, VisualItem.Y2, Y);
    }
    

    /**
     * creates a bounding box for the station in the "from" column of a visual item.
     * @param item the visualitem for that the bounding box is created
     * @return the bounding box of the station
     */
    private Rectangle getFromStationBounds(VisualItem item){
        Rectangle rect = new Rectangle(stationBounds);
        Table t = item.getTable();
                
        if(item.getRow() == 0 || !item.getString("connectionID").equals(t.getString(item.getRow() - 1, "connectionID"))){
            rect.x = ((int) item.getX()) - rect.width / 2;
            rect.y = ((int) item.getY()) - rect.height / 2;
        }else{
            VisualItem nnn = (VisualItem) t.getTuple(item.getRow() - 1);
            Line2D line = (Line2D) trainRenderer.getTrainLine(nnn);
            
            rect.x = (int) (line.getX2() + (item.getX() - line.getX2()) / 2);
            rect.y = (int) item.getY();
            
            rect.x = rect.x - rect.width / 2;
            rect.y = rect.y - rect.height / 2;
        }
        
        return rect;
    }
    
    /**
     * Sets the tooltip of a given display to an infobox about a specific station.
     * @param to the train that has the desired station as destination-station.
     * @param from the train that has the desired station as departure-station.
     * @param d the tooltip of this display is set.
     */
    private void setToolTipforStation(VisualItem to, VisualItem from, Display d){
        String TipText = "";
        if(to == null){
            String stationInfo = stationsTable.getString(searchTable(Integer.toString(from.getInt("from")), "stationKM", stationsTable), "stationInfo");
            stationInfo = stationInfo.replaceAll("~", "<br> - ");
            SimpleDateFormat dateFormat = new SimpleDateFormat( "HH:mm" );
            TipText = 
                    "<HTML>" +
                    "<span style=\"font-family:arial\">" +
                    "<b>" +
                    "<font size=\"+1\">" +
                    stationsFormat.format(from.getInt("from")) +
                    "</font>" +
                    "<br>" +
                    "&nbsp;&nbsp;" +
                    Functions.getFieldforID(from.getString("trainID"), "trainID", "trainName", trainsTable) +
                    "</b>" +
                    "<br>" +
                    "<hr>" +
                    "- Ab: " + dateFormat.format(from.getDate("fromTime")) +
                    "<hr>" +
                    "  - " +
                    stationInfo +
                    "</span>" +
                    "</HTML>";
        }else if(from == null){
            String stationInfo = stationsTable.getString(searchTable(Integer.toString(to.getInt("to")), "stationKM", stationsTable), "stationInfo");
            stationInfo = stationInfo.replaceAll("~", "<br> - ");
            SimpleDateFormat dateFormat = new SimpleDateFormat( "HH:mm" );
            TipText = 
                    "<HTML>" +
                    "<span style=\"font-family:arial\">" +
                    "<b>" +
                    "<font size=\"+1\">" +
                    stationsFormat.format(to.getInt("to")) +
                    "</font>" +
                    "<br>" +
                    "&nbsp;&nbsp;" +
                    Functions.getFieldforID(to.getString("trainID"), "trainID", "trainName", trainsTable) +
                    "</b>" +
                    "<br>" +
                    "<hr>" +
                    "- An: " + dateFormat.format(to.getDate("toTime")) +
                    "<hr>" +
                    "  - " +
                    stationInfo +
                    "</span>" +
                    "</HTML>";
        }else{
            String stationInfo = stationsTable.getString(searchTable(Integer.toString(from.getInt("from")), "stationKM", stationsTable), "stationInfo");
            stationInfo = stationInfo.replaceAll("~", "<br> - ");
            SimpleDateFormat dateFormat = new SimpleDateFormat( "HH:mm" );
            TipText = 
                    "<HTML>" +
                    "<span style=\"font-family:arial\">" +
                    "<b>" +
                    "<font size=\"+1\">" +
                    stationsFormat.format(from.getInt("from")) +
                    "</font>" +
                    "<br>" +
                    "&nbsp;&nbsp;" +
                    Functions.getFieldforID(to.getString("trainID"), "trainID", "trainName", trainsTable) +
                    "</b>" +
                    "<br>" +
                    "<hr>" +
                    "- An: " + dateFormat.format(to.getDate("toTime")) +
                    "<br>" +
                    "- Ab: " + dateFormat.format(from.getDate("fromTime")) +
                    "<hr>" +
                    "  - " +
                    stationInfo +
                    "</span>" +
                    "</HTML>";
        }
        
                
        d.setToolTipText(TipText);        
    }
    
    /**
     * Highlights the station that the cursor currently points to. Also sets the tooltip for a given display.
     * @param e the mouse event for the current mouseposition.
     * @param d the tooltip of this display is set to a informationbox about the current station.
     * @return the train that contains the currently selected station.
     */
    private VisualItem highlightSelectedStation(MouseEvent e, Display d){
                Iterator it = MyVis.visibleItems("data");
                Rectangle bound = new Rectangle(stationBounds);
                VisualTable t = (VisualTable) MyVis.getGroup("data");
                
                while(it.hasNext()){
                    VisualItem act = (VisualItem) it.next();
                    
                    if(act.get("fromPoint") != null){
                        bound.x = ((Point) act.get("fromPoint")).x;                        
                        bound.y = ((Point) act.get("fromPoint")).y;
                        bound.x = bound.x - bound.width / 2;
                        bound.y = bound.y - bound.height / 2;

                        if(bound.contains(e.getPoint())){
                            setHighlight1(bound.x + bound.width / 2, bound.y + bound.height / 2);
                            setToolTipforStation(t.getItem(act.getRow()-1), act, d);
                            return act;
                        }else setHighlight1(-100, -100);
                    }else{
                        bound.x = (int) act.getX() - bound.width / 2;
                        bound.y = (int) act.getY() - bound.height / 2;
                        
                        
                        if(bound.contains(e.getPoint())){
                            setHighlight1(bound.x + bound.width / 2, bound.y + bound.height / 2);
                            setToolTipforStation(null, act, d);
                            return act;
                        }else setHighlight1(-100, -100);
                    }
                    
                    if(act.get("toPoint") != null){
                        bound.x = ((Point) act.get("toPoint")).x;
                        bound.y = ((Point) act.get("toPoint")).y;
                        bound.x = bound.x - bound.width / 2;
                        bound.y = bound.y - bound.height / 2;

                        if(bound.contains(e.getPoint())){
                            setHighlight1(bound.x + bound.width / 2, bound.y + bound.height / 2);
                            setToolTipforStation(act, t.getItem(act.getRow()+1), d);
                            return act;
                        }else setHighlight1(-100, -100);
                    }else{
                        Line2D line = (Line2D) trainRenderer.getTrainLine(act);
                        
                        bound.x = (int) line.getX2() - bound.width / 2;
                        bound.y = (int) line.getY2() - bound.height / 2;                        
                        
                        if(bound.contains(e.getPoint())){
                            setHighlight1(bound.x + bound.width / 2, bound.y + bound.height / 2);
                            setToolTipforStation(act, null, d);
                            return act;
                        }else setHighlight1(-100, -100);
                    }
                    
                }
                
                return null;
    }
    
    /**
     * Creates a new instance of LineVis.
     * Most of the classes work is done here.
     * @param connections table of the connections to be visualized
     * @param stations table of the stations a train can stop at. Affects the size and labeling of the Y axis.
     * @param trains table of the trains that are used in the connections
     */
    public LineVis(Table connections, Table stations, Table trains, Table dates, Vector _selectedTrains, JTabbedPane _tabPane) {
        super(new BorderLayout());
        
        tabPane = _tabPane;
        datesTable = dates;
        selectedTrains = _selectedTrains;
        //save the "database" to the class:
        connectionsTable = Functions.makeData(connections, stations, trains, dates);
        tabPane.firePropertyChange("_PUSHPROGRESS_", 0, 7);
        changeStationIDtoStationKM(connectionsTable, stations);
        connectionsTable.addColumn("fromPoint", Point.class, null);
        connectionsTable.addColumn("toPoint", Point.class, null);
                
        stationsTable = stations;
        trainsTable = trains;
        
        double[] stationKMs = new double[stations.getRowCount()];
                for(int i = 0; i<stationKMs.length; i++)
                    stationKMs[i] = stations.getInt(i, "stationKM"); //StationIDs are supposed to start form 1 
        
            String[] stationNames = new String[stationKMs.length];
                for(int j = 0; j<stationNames.length; j++)
                    stationNames[j] = stations.get(j, "stationName").toString();
        
            //set the ChoiceFormat for the mapping of IDs to names
        stationsFormat = new ChoiceFormat(stationKMs, stationNames);
        
        d = new LineVisDisplay(connectionsTable, stationsTable);
        MyVis = d.getVisualization();
        tabPane.firePropertyChange("_PUSHPROGRESS_", 0, 7);
        
        //create a schema for the highlight table
        Schema s =  PrefuseLib.getVisualItemSchema();
        s.addColumn(VisualItem.X2, double.class);
        s.addColumn(VisualItem.Y2, double.class);
        
        //create the table and add 2 rows for 2 highlights
        Table highlight = s.instantiate();
        highlight.addRow();
        highlight.addRow();
        
        //set both highlights to "invisible"
        Tuple tuple = highlight.getTuple(0);
        tuple.set(VisualItem.X2, -100.0);
        tuple.set(VisualItem.Y2, -100.0);
        
        tuple = highlight.getTuple(1);
        tuple.set(VisualItem.X2, -100.0);
        tuple.set(VisualItem.Y2, -100.0);
        
        MyVis.addTable("highlight", highlight);
        tabPane.firePropertyChange("_PUSHPROGRESS_", 0, 7);
        
        
        //set a ControlAdapter to set the actions to be taken if the mouse is moved over a VisualItem
        Control hoverc = new ControlAdapter() {
            VisualItem lastitem = null;
            public void mouseMoved(MouseEvent e){
                VisualItem selected = highlightSelectedStation(e, d);
                if(selected != null){
                    highlightTrain(selected.getDouble("ID"), highlightColorInt);
                    lastitem = selected;
                }else{
                    if(lastitem != null){
                        dehighlightTrain(lastitem.getDouble("ID"));
                    
                    setHighlight1(-100, -100);

                    //set the color of the visualitems back to normal
                    dehighlightTrain(lastitem.getDouble("ID"));
                  
                    d.setToolTipText(null);
          
                    //repaint is needed because of the colorchange
                    lastitem.getVisualization().repaint();
                }
                    lastitem = null;
                }
            }
            
            public void itemEntered(VisualItem item, MouseEvent e){
                if(item.isInGroup("data")){
                    if(lastitem != null) 
                        if(item.getDouble("ID") != lastitem.getDouble("ID")) dehighlightTrain(lastitem.getDouble("ID"));
                    
                    itemMoved(item, e);
                }
            }
            
            public void itemMoved(VisualItem item, MouseEvent e){
                if(item.isInGroup("data")){
                    lastitem = item;

                    //set the color of all items with the selected trainID to red
                    highlightTrain(item.getDouble("ID"), highlightColorInt);
                    
                    String trainInfo = Functions.getFieldforID(item.getString("trainID"), "trainID", "trainInfo", trainsTable).toString();
                    trainInfo = trainInfo.replaceAll("~", "<br> -");
                    trainInfo = trainInfo.substring(0, trainInfo.length()-2);

                    
                    SimpleDateFormat dateFormat = new SimpleDateFormat( "HH:mm" );
                    String tipText =
                            "<HTML>" +
                            "<span style=\"font-family:arial\">" +
                            "<b>" +
                            "<font size=\"+1\">" +
                            Functions.getFieldforID(item.getString("trainID"), "trainID", "trainName", trainsTable) +
                            "</font>" +
                            "</b>" +
                            "&nbsp;&nbsp;&nbsp;" +
                            dateFormat.format(connectionsTable.get(searchTable(item.getString("ID"), "ID", connectionsTable), "fromTime")) +
                            " - " +
                            dateFormat.format(connectionsTable.get(Functions.searchLast("ID", item.getString("ID"), connectionsTable), "toTime")) +
                            "<br>" +
                            "<b>" +
                            "&nbsp;&nbsp;" +
                            stationsFormat.format(connectionsTable.get(searchTable(item.getString("ID"), "ID", connectionsTable), "from")) + 
                            " - " +
                            stationsFormat.format(connectionsTable.get(Functions.searchLast("ID", item.getString("ID"), connectionsTable), "to")) + 
                            "</b>" +
                            "</span><br>" +
                            "<hr>" +
                            "  - " +
                            trainInfo +
                            "<hr>" + 
                            "&nbsp;&nbsp;" +
                            Functions.trainOnDateInfo(item.getString("connectionID"), datesTable) +
                            "</span>" +
                            "</HTML>";
                    
                    d.setToolTipText(tipText);
                    
                    highlightSelectedStation(e, d);
                    
                    
                    //repaint is needed because of the colorchange
                    item.getVisualization().repaint();
                    
                }
            }
            
            public void itemClicked(VisualItem item, MouseEvent e){
                if (!SwingUtilities.isLeftMouseButton(e)) return;
                if(item.isInGroup("highlight")) item = lastitem;
                if (!item.isInGroup("data")) return;
                if ( e.getClickCount() == 1 ) {
                    selectTrain(item.getString("trainID"), !item.isHighlighted());
                    if (selectedTrains.contains(item.getString("trainID"))) selectedTrains.remove(item.getString("trainID"));
                    else selectedTrains.add(item.getString("trainID"));
                
                    // fire PropertyChange on TabbedPane, the container of oll visualisations
                    d.getParent().getParent().getParent().firePropertyChange("_REPAINT_", 0, 1);
                }
                
            }

        };
        d.addControlListener(hoverc);
        
        addPropertyChangeListener(new PropertyChangeListener() {
            private boolean addBorder = false;
            public void propertyChange(PropertyChangeEvent evt) {
                
                if (evt.getPropertyName().equals("_REPAINT_")) {
                    MyVis.run("update");
                }
                
                if (evt.getPropertyName().equals("_SETSORT_")) {
                    
                    int sort = (int)Functions.object2Long(evt.getNewValue());
                    if(sort == 0) UPorDOWN = true;
                    if(sort == 1) UPorDOWN = false;
                    MyVis.run("layout");
                }
                
                if (evt.getPropertyName().equals("_FISHEYE_")) { 
                    
                    double fisEye = (double)Functions.object2Long(evt.getNewValue());
                    d.timeScale.setFisheyeIntensity(fisEye/1000);
                }
                
                if (evt.getPropertyName().equals("_STARTDATE_")) {
                    
                    if(Functions.object2Long(evt.getOldValue()) == 1337) return;
                    long newVal = Functions.object2Long(evt.getNewValue());
                    long endDate = d.timeScale.getEndDate();
                    
                    d.timeScale.changeInterval(
                        newVal - d.timeScale.getMillisPerPixel()*(long)yAxisBorder,
                        endDate
                        );
                }
                
                if (evt.getPropertyName().equals("_ENDDATE_")) {
                    
                    long startDate = d.timeScale.getStartDate();
                    d.timeScale.changeInterval(
                            startDate,
                            Functions.object2Long(evt.getNewValue())                            
                            );
                }
                
                if (evt.getPropertyName().equals("_SETDIRECTION_")) { 
                    
                    int direction = (int)Functions.object2Long(evt.getNewValue());
                    d.dir.setDirection(direction);
                    MyVis.run("update");
                }    
                

                
                if (evt.getPropertyName().equals("_SHOWSTATIONS_")) { 
                    
                    long newVal = Functions.object2Long(evt.getNewValue());
                    trainRenderer.drawStations(newVal == 1 ? true : false);
                    MyVis.run("update");
                }
            }
        });

        
        JPanel JPanelVis = new JPanel();
        JPanelVis.setLayout(new java.awt.BorderLayout());
        JPanelVis.add(d, BorderLayout.CENTER);
        
        TimeScaleHeader tsh = new TimeScaleHeader(d.timeScale);
        
        JPanelVis.add(tsh, BorderLayout.NORTH);
        
        StatusBar statusBar = new StatusBar(d.timeScale);
        statusBar.add(Box.createHorizontalStrut(20));
        JPanelVis.add(statusBar, BorderLayout.SOUTH);
        JPanelVis.add(makeJRangeSlider_Y(), BorderLayout.EAST);

        add(JPanelVis, BorderLayout.CENTER);

        
        MyVis.run("layout");

    }
    
    /**
     * The Display-class for LineVis. It creates the renderer as well as all necessary actions. It also holds functionality for panning and zooming and therefore manages the timescale-object of this visualization.
     */
    private class LineVisDisplay extends Display {
        
        private static final String DATA = "data";
        
        /**
         * the timescale of this visualization.
         */
        private FisheyeTimeScale timeScale = null;
        /**
         * This is used to draw the timescale in the back of the panel
         */
        private LineTimeScalePainter timeScalePainter = null;
        /**
         * this draws a line at the position of the mouse and also shows the time at that position.
         */
        private MouseTracker mouseTracker = null;
        /**
         * the Layout for the y-axis and its labels.
         */
        private stationAxisLabelLayout yAxis = null;
        /**
         * 
         */
        private Direction dir = null;
        
        private AxisLayout toLay;
        private AxisLayout fromLay;
        
        /**
         * creates a tooltip for this display.
         */
        public JToolTip createToolTip() {
            
            JToolTip jToolTip = new JToolTip();
            jToolTip.setBackground(Color.LIGHT_GRAY);
            
            return jToolTip;
        }
        
        /**
         * pans the timescale.
         * @param dx the difference to the old x coordinate.
         * @param dy in our application only horizontal panning is allowed. Therefore this parameter is ignored.
         */
        public synchronized void pan(double dx, double dy) {
            panAbs(dx, dy);
        }
        
        /**
         * pans the timescale.
         * @param dx the difference to the old x coordinate.
         * @param dy in our application only horizontal panning is allowed. Therefore this parameter is ignored.
         */
        public synchronized void panAbs(double dx, double dy) {
            if (timeScale != null) {
                
                long oldStartDate = timeScale.getStartDate();
                long oldEndDate = timeScale.getEndDate();
                
                long diffMillis = timeScale.getMillisPerPixel() * (int)dx;
                
                long newStartDate = oldStartDate - diffMillis;
                long newEndDate = oldEndDate - diffMillis;
                
                timeScale.setStartDate(newStartDate);
                
                d.getParent().getParent().getParent().firePropertyChange(
                        "_STARTDATE_", 1337, newStartDate
                        );
                
                timeScale.setEndDate(newEndDate);
                
                d.getParent().getParent().getParent().firePropertyChange(
                        "_ENDDATE_", oldEndDate, newEndDate
                        );
            }
        }
        
        /**
         * zooms the timescale
         * @param scale the new scalefactor
         */
        public synchronized void zoom(Point2D p, double scale) {
            zoomAbs(p, scale);
        }
        
        /**
         * zooms the timescale
         * @param scale the new scalefactor
         */
        public synchronized void zoomAbs(Point2D p, double scale) {
            if (timeScale != null) {
                
                Date oldStartDate = new Date(timeScale.getStartDate());
                Date oldEndDate = new Date(timeScale.getEndDate());
                
                timeScale.zoom(scale);
                
                Date newStartDate = new Date(timeScale.getStartDate());
                Date newEndDate = new Date(timeScale.getEndDate());
                
                d.getParent().getParent().getParent().firePropertyChange(
                        "_STARTDATE_",
                        1337,
                        newStartDate.getTime()
                        );
                d.getParent().getParent().getParent().firePropertyChange(
                        "_ENDDATE_",
                        oldEndDate.getTime(),
                        newEndDate.getTime()
                        );
            }
        }
        
        /**
         * creates a new instance of LineVisDisplay
         * @param data the table that holds all connections.
         * @param stations the table that holds information about the stations
         */
        public LineVisDisplay(Table data, Table stations){
            super(new Visualization());

            setHighQuality(true);
            
            GregorianCalendar d = new GregorianCalendar();
            
            timeScale = new FisheyeTimeScale(Functions.getEarliestDate(data, d).getTime().getTime()-1000*60*30, Functions.getLatestDate(data, d).getTime().getTime()+1000*60*30, getWidth() - 1);

            
            timeScale.setAdjustDateRangeOnResize(false);
            timeScale.addChangeListener(new ChangeListener() {
                public void stateChanged(ChangeEvent e) {
                    getVisualization().run("update");
                }
            });
            
            addComponentListener(new ComponentAdapter() {
                public void componentResized(ComponentEvent e) {
                    timeScale.setDisplayWidth(getWidth() - 1);
                    getVisualization().run("layout");
                }
            });
            
            addControlListener(new PanControl());
            addControlListener(new ZoomControl());

            
            timeScalePainter = new LineTimeScalePainter(this, yAxisBorder);
            timeScalePainter.setTimeScale(timeScale);
            
            mouseTracker = new MouseTracker(this, timeScale);
            
        
            trainRenderer = new LineRenderer(DATA, getVisualization());
            
            trainRenderer.setTimeScale(timeScale);
    
            RendererFactory ref = new RendererFactory() {
                AxisRenderer arY = new AxisRenderer(Constants.LEFT, Constants.TOP);

            
                public Renderer getRenderer(VisualItem item) {
                    return item.isInGroup("ylab") ? arY : 
                            item.isInGroup("highlight") ? dotrenderer : trainRenderer;
                }
            };
            getVisualization().setRendererFactory(ref);
            
            
            addPaintListener(new PaintListener() {
                public void postPaint(Display d, Graphics2D g) {
                    mouseTracker.paintTimeAtPosition(g);
                    g.setColor(new Color(100, 100, 100));
                    g.fillRect(yAxisBorder - 1, 0, 2, getHeight());
                }
                
                public void prePaint(Display d, Graphics2D g) {
                    g.transform(d.getInverseTransform());
                    timeScalePainter.paint(g);
                    g.transform(d.getTransform());
                }
            });
            
            
            
            data.addColumn("toY", double.class, 0);
            getVisualization().addTable(DATA, data);
            tabPane.firePropertyChange("_PUSHPROGRESS_", 0, 7);
            
            NumberRangeModel stationRange = new NumberRangeModel(0, stations.getInt(stations.getRowCount()-1, "stationKM"), 0, stations.getInt(stations.getRowCount()-1, "stationKM"));
            toLay = new AxisLayout(DATA, "to", Constants.Y_AXIS);
            fromLay = new AxisLayout(DATA, "from", Constants.Y_AXIS);
            toLay.setRangeModel(stationRange);
            fromLay.setRangeModel(stationRange);
            
            
            dir = new Direction("data", Direction.UP);
                        
            getVisualization().addTable("ylab", PrefuseLib.getAxisLabelSchema());
            yAxis = new stationAxisLabelLayout("ylab", stations);
            
            ActionList layout = new ActionList();
            layout.add(new ColorAction(DATA, VisualItem.FILLCOLOR, ColorLib.rgb(0,0,0)));
            layout.add(new ColorAction(DATA, VisualItem.STROKECOLOR, trainColor));
            layout.add(toLay);
            layout.add(new YtoEndY(DATA));
            layout.add(fromLay);
            layout.add(new TimeLayout(DATA, "fromTime", timeScale));
            layout.add(yAxis);
            layout.add(new BorderFilter(DATA));
            layout.add(dir);
            layout.add(new SelectTrains(DATA));
            layout.add(new FlipDisplay(DATA));
            layout.add(new RepaintAction());
            getVisualization().putAction("layout", layout);
            
        
            ActionList update = new ActionList();
            
            update.add(new TimeLayout(DATA, "fromTime", timeScale));
            update.add(yAxis);
            update.add(new BorderFilter(DATA));
            update.add(new SelectTrains(DATA));
            update.add(dir);
            update.add(new RepaintAction());
            getVisualization().putAction("update", update);
            
            
            //Order of the items to be drawn:
            //first the 2 axis, then the trains(lines) and at the end the highlights
            setItemSorter(new ItemSorter() {
                public int score(VisualItem item) {
                    if(item.isInGroup(DATA)){ return 2; }
                        else if (item.isInGroup("highlight")){ return 3; }
                    else return 1; 
                }
            });
            
            addComponentListener(new ComponentAdapter() {
                public void componentResized(ComponentEvent e) {
                    timeScale.setDisplayWidth(getWidth() - 1);
                }
            });
            
        }
                
    }
}
