/*
 * Main.java
 *
 * Created on April 29, 2007, 9:08 PM
 *
 */

package zugvis;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.GregorianCalendar;
import java.util.Vector;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import prefuse.util.ui.UILib;
import javax.swing.JFrame;
import javax.swing.JTabbedPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.UIManager;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;
import javax.swing.JOptionPane;

import prefuse.data.Table;
import prefuse.data.io.DelimitedTextTableReader;

/**
 * The Main class of the application. The frame and all components are instatiated here. Also all datafiles are loaded and tested here.
 * @author nice
 */
public class Main {
    
    /**
     * Maximum size of zipfile for loading tables
     */
    final static int ZIPFILE_MAX_SIZE = 1000000;
    
    /**
     * Contains the id's of selected trains
     */
    private static Vector selectedTrains = new Vector();
    
    /**
     * Contains the zipfilename of tables
     */
    private static Vector tablesZipFile = new Vector();
    
    /**
     * Contains the string of os-specific fileseparator
     */
    private static String fileSep = System.getProperty("file.separator");
    
    /**
     * Tempdirectory to extract tables of loadet zipfile
     */
    private static String tempDir = "_temp_tables";
    
    /**
     * Remember last seted startdate to update visualisations on tabchange
     */
    private static long startDate = 0;
    
    /**
     * Remember previous seted startdate to update visualisations on tabchange
     */
    private static long oldstartDate = 0;
    
    /**
     * Remember last seted enddate to update visualisations on tabchange
     */
    private static long endDate = 10;
    
    /**
     * Remember previous seted enddate to update visualisations on tabchange
     */
    private static long oldendDate = 10;
    
    /**
     * Remember fisheyeintensity to update visualisations on tabchange
     */
    private static long fishEye = 0;
    
    /**
     * Remember direction to update visualisations on tabchange
     */
    private static long direction = 0;
    
    /**
     * Tabname of textvisualisation
     */
    private static String TEXTVIS = "Table";
    
    /**
     * Tabname of barchartvisualisation relative
     */
    private static String BARCHARTVISREL = "Duration Bar Chart";
    
    /**
     * Tabname of bachartvisualisation absolute
     */
    private static String BARCHARTVISABS = "Time Line Chart";
    
    /**
     * Tabname of linevisualisation
     */
    private static String LINEVIS = "Marey Diagram";
    
    /**
     * Tabname of errorpane
     */
    private static String ERROR = "Error";
    
    /**
     * Tabname of helppane
     */
    private static String HELP = "Help";
    
    /**
     * Tabname of loadingpane
     */
    private static String LOADING = "Loading tables...";
    
    /**
     * Width of helppane in pixel
     */
    private static int HELP_WIDTH = 600;
    
    /**
     * Height of helppane in pixel
     */
    private static int HELP_HEIGHT = 800;
    
    /**
     * Number of connections
     */
    private static int numberConnections = 0;
    
    static JFrame frame = null;
    
    static TextVis textVis = null;
    static BarChartVisAbs barChartVisAbs = null;
    static BarChartVisRel barChartVisRel = null;
    static LineVis lineVis = null;
    
    static JTextPane errorPane = new JTextPane();
    static JScrollPane errorScrollPane = new JScrollPane();
    static JTextPane helpPane = new JTextPane();
    static Color helpPaneColour = new Color(204,204,204);
    static JScrollPane helpScrollPane = new JScrollPane();
    static JPanel jPanelLoading = new JPanel();
    static JProgressBar progressBar = new JProgressBar();
    static JLabel progressBarLabel = new JLabel();
    
    static JFrame HelpFrame = new JFrame("Help");
    
    static JMenuBar menuBar;
    static JMenu File;
    static JMenu Direction;
    static JMenu Sort;
    static JMenu Info;
    static JMenuItem Load;
    static JMenuItem Exit;
    static JMenuItem About;
    static JMenuItem Help;
    static JCheckBoxMenuItem fromto;
    static JCheckBoxMenuItem tofrom;
    static JRadioButtonMenuItem sortUP;
    static JRadioButtonMenuItem sortDOWN;
        
    static Menu menu;
    
    static int progressCounter = 0;
    static String errorMsg = "";
    
    /**
     * Initialize tabbed pane (extract zip, check and load tables)
     * @param tabPane Tabpane to initialize
     * @param menu Toolbar for vizualizationcommands
     * @param tempDir Tempdirectory to extract zipfile of tables
     */
    private static void makeTabbedPane(final JTabbedPane tabPane, final Menu menu, final String tempDir) {
        
        UILib.setPlatformLookAndFeel();
        
        tabPane.remove(textVis);
        tabPane.remove(barChartVisAbs);
        tabPane.remove(barChartVisRel);
        tabPane.remove(lineVis);
        tabPane.remove(errorScrollPane);
        tabPane.remove(helpScrollPane);
        
        progressBar.setValue(progressCounter = 0);
        progressBarLabel.setText("Loading...");
        tabPane.add(jPanelLoading, LOADING);
        
        errorMsg = "";
        
        // Thread for loading tables and initialize visualizations
        Thread runner = new Thread() {
            
            private void getError(String message) {
                
                progressBarLabel.setForeground(Color.RED.darker());
                progressBarLabel.setText(message);
                if (!message.equals("Error while loading!")) errorMsg += message + "~";
                
                try { Thread.sleep(2500); } catch (Exception ex) { }
                
                progressBarLabel.setText("");
                progressBarLabel.setForeground(new Color(8, 32, 128));
            }
            
            public void run() {
                
                int errors = 0;
                
                for (int i=0; i<1; i++) {
                    
                    // Start und Enddatum initialisieren
                    GregorianCalendar tempD = new GregorianCalendar();
                    
                    tempD.set(tempD.get(tempD.YEAR), tempD.get(tempD.MONTH), tempD.get(tempD.DAY_OF_MONTH), 0, 0, 0);
                    startDate = tempD.getTimeInMillis();
                    
                    tempD.set(tempD.get(tempD.YEAR), tempD.get(tempD.MONTH), tempD.get(tempD.DAY_OF_MONTH), 23, 59, 59);
                    endDate = tempD.getTimeInMillis();
                    
                    progressBarLabel.setText("Loading zipfile");
                    progressBar.setValue(progressCounter = 2);
                    
                    Table connectionsTable = null, stationsTable = null, trainsTable = null, datesTable = null;
                    
                    // read Zip with tables
                    errorMsg = Functions.extractTables((String)tablesZipFile.get(0), tempDir);
                    if (!errorMsg.equals("")) { errors++; break; }
                    
                    
                    try {
                        connectionsTable = new DelimitedTextTableReader(";").readTable(tempDir + fileSep + "connections.csv");
                    } catch ( Exception e ) { getError("Error reading \"connections.csv\""); errors++; }
                    
                    try {
                        stationsTable = new DelimitedTextTableReader(";").readTable(tempDir + fileSep + "stations.csv");
                    } catch ( Exception e ) { getError("Error reading \"stations.csv\""); errors++; }
                    
                    try {
                        trainsTable = new DelimitedTextTableReader(";").readTable(tempDir + fileSep + "trains.csv");
                    } catch ( Exception e ) { getError("Error reading \"trains.csv\""); errors++; }
                    
                    try {
                        datesTable = new DelimitedTextTableReader(";").readTable(tempDir + fileSep + "dates.csv");
                    } catch ( Exception e ) { getError("Error reading \"dates.csv\""); errors++; }
                    
                    Functions.deleteDirectory(new File(tempDir));
                    
                    if (errors != 0) { break; }
                    
                    progressBar.setValue(progressCounter = 10);
                    progressBarLabel.setText("Validating tables");
                    
                    errorMsg = Functions.validateTables(connectionsTable, stationsTable, trainsTable, datesTable);
                    if (!errorMsg.equals("")) { errors++; break; }
                    
                    progressBar.setValue(progressCounter = 15);
                    
                    progressBarLabel.setText("Loading " + TEXTVIS);
                    try {
                        textVis = new TextVis(connectionsTable, stationsTable, trainsTable, datesTable);
                        tabPane.add(TEXTVIS, textVis);
                    } catch ( Exception e ) { getError("Error while loading " + TEXTVIS); errors++; }
                    progressBar.setValue(progressCounter = 20);
                    
                    progressBarLabel.setText("Loading " + BARCHARTVISABS);
                    try {
                        barChartVisAbs =  new BarChartVisAbs(tabPane, selectedTrains, connectionsTable, stationsTable, trainsTable, datesTable);
                        tabPane.add(BARCHARTVISABS,barChartVisAbs);
                    } catch ( Exception e ) { getError("Error while loading " + BARCHARTVISABS); errors++; }
                    progressBar.setValue(progressCounter = 50);
                    
                    progressBarLabel.setText("Loading " + BARCHARTVISREL);
                    try {
                        barChartVisRel = new BarChartVisRel(tabPane, selectedTrains, connectionsTable, stationsTable, trainsTable, datesTable);
                        tabPane.add(BARCHARTVISREL, barChartVisRel);
                    } catch ( Exception e ) { getError("Error while loading " + BARCHARTVISREL); errors++; }
                    progressBar.setValue(progressCounter = 65);
                    
                    progressBarLabel.setText("Loading " + LINEVIS);
                    try {
                        lineVis = new LineVis(connectionsTable, stationsTable, trainsTable, datesTable, selectedTrains, tabPane);
                        tabPane.add(LINEVIS, lineVis);
                    } catch ( Exception e ) { getError("Error while loading " + LINEVIS); errors++; }
                    progressBar.setValue(progressCounter = 98);
                     
                    fromto.setText(stationsTable.getString(0, "stationName") + " - " + stationsTable.getString(stationsTable.getRowCount()-1, "stationName"));
                    tofrom.setText(stationsTable.getString(stationsTable.getRowCount()-1, "stationName") + " - " + stationsTable.getString(0, "stationName"));
                    
                    sortDOWN.setText(stationsTable.getString(0, "stationName") + " - " + stationsTable.getString(stationsTable.getRowCount()-1, "stationName"));
                    sortUP.setText(stationsTable.getString(stationsTable.getRowCount()-1, "stationName") + " - " + stationsTable.getString(0, "stationName"));


                    numberConnections = connectionsTable.getInt(connectionsTable.getRowCount()-1, "connectionID");
                    
                    progressBarLabel.setText("Loading " + HELP);
                    loadHelpFile();
                    if (helpPane != null) HelpFrame.add(helpScrollPane);

                    progressBar.setValue(progressCounter = 98);
                    
                    progressBarLabel.setText("Setting menubar");
                    menu.setDates(datesTable);
                    progressBar.setValue(progressCounter = 100);
                } // end of for (1)
                
                if (errors == 0) {
                    
                    progressBarLabel.setText("Loading complete.");
                    try { Thread.sleep(700); } catch (Exception ex) { }
                    tabPane.remove(jPanelLoading);
                    setMenu(tabPane, menu);
                } else {
                    
                    progressCounter = 100;
                    
                    getError("Error while loading!");
                    
                    errorMsg = errorMsg.replaceAll("~", "<br>");
                    if (errorMsg.endsWith("<br>")) errorMsg = errorMsg.substring(0, errorMsg.length()-4);
                    
                    errorPane.setText(
                            "<table align=\"center\" style=\"font-family: arial;\">" +
                            "<tr>" +
                            "<td align=\"center\">" +
                            "<h3>" +
                            "Errors while loading new visualizations:</td>" +
                            "</h3>" +
                            "</tr>" +
                            "<tr>" +
                            "<td align=\"left\" style=\"color: red;\">" + errorMsg + "</td>" +
                            "</tr>" +
                            "</table>"                            
                            );
                    
                    errorScrollPane.setViewportView(errorPane);
                    tabPane.add(ERROR, errorScrollPane);
                    
                    tabPane.remove(helpScrollPane);
                    if (helpPane != null) tabPane.add(HELP, helpScrollPane);
                    
                    tabPane.remove(jPanelLoading);
                    setMenu(tabPane, menu);
                }
            }
        };
        runner.start();
    }
    
    /**
     * Loads the help.html into the helppane
     */
    private static void loadHelpFile() {
        
        try {
            
            File helpFile = new File("help.htm");
            
            InputStream fis = null;
            if (helpFile.exists()) fis = new FileInputStream(helpFile.getName());
            else fis = Functions.class.getClassLoader().getResourceAsStream(helpFile.getName());
           
            int x= fis.available();
            byte b[]= new byte[fis.available()];
            fis.read(b);
            helpPane.setText(new String(b));
            helpPane.setBackground(helpPaneColour);
        } catch ( Exception e ) {
            
            helpPane = null;
            System.err.println("Error reading \"help.htm\"");
        }
        
        helpScrollPane.setViewportView(helpPane);
        
    }
    
    /**
     * Initzialize the Toolbar for visualizationcommands
     * @param tabPane Tabpane of visualizations
     * @param menu Toolbar for visualizationcommands
     */
    private static void setMenu(JTabbedPane tabPane, Menu menu) {
        
        if (tabPane.getSelectedIndex() >= 0) {
            
            String tabTitle = tabPane.getTitleAt(tabPane.getSelectedIndex());
            
            if ( tabTitle.equals(TEXTVIS) ) {
                menuBar.setEnabled(true);
                menu.setAllEnabled(true); 
                menu.setTimeControlsEnabled(false); 
                menu.setComponetEnabled("STATIONSCHECKBOX", false);
                menu.setComponetEnabled("FITTIMEBUTTON", false); 
                Sort.setEnabled(false);
            } else if ( tabTitle.equals(BARCHARTVISREL) ) {
                menuBar.setEnabled(true);
                menu.setAllEnabled(false); 
                menu.setComponetEnabled("STATIONSCHECKBOX", true); 
                Sort.setEnabled(false);
            } else if ( tabTitle.equals(BARCHARTVISABS) ){
                menuBar.setEnabled(true);
                menu.setAllEnabled(true);
                Sort.setEnabled(false);
            }
            else if ( tabTitle.equals(LINEVIS) ) {
                menuBar.setEnabled(true);
                menu.setAllEnabled(true);
                Sort.setEnabled(true);
            }
            else if ( tabTitle.equals(LOADING) ) {
                menuBar.setEnabled(false);                
                menu.setAllEnabled(false);
            }
            else if ( tabTitle.equals(ERROR) ) {
                menu.setAllEnabled(false);
                menuBar.setEnabled(false);
            }
            else {
                menuBar.setEnabled(false);
                menu.setAllEnabled(false);
            }
        }
    }
    
    
    
    /**
     * Mainfunction
     * @param args Commandline parameters
     */
    public static void main(String[] args) {
        
        UILib.setPlatformLookAndFeel();
        UIManager.put("ProgressBar.selectionBackground", Color.black);
        UIManager.put("ProgressBar.selectionForeground", Color.white);
        UIManager.put("ProgressBar.foreground", new Color(8, 32, 128));
        
        tablesZipFile.add("tables.zip");
        
        helpPane.setContentType("text/html");
        errorPane.setContentType("text/html");
        
        helpPane.setEditable(false);
        errorPane.setEditable(false);
        
        helpPane.setSize(new Dimension(700, 400));
        errorPane.setSize(new Dimension(700, 400));
        
        errorPane.setBackground(new JLabel().getBackground());        
        
        progressBar.setMinimum(0);
        progressBar.setMaximum(100);
        progressBar.setStringPainted(true);
        progressBar.setSize(new Dimension(300, 40));
        
        jPanelLoading.setLayout(new BoxLayout(jPanelLoading, BoxLayout.X_AXIS));
        jPanelLoading.setPreferredSize(new Dimension(700, 400));
        jPanelLoading.setBackground(new JLabel().getBackground());
        jPanelLoading.add(Box.createHorizontalGlue());
        
        JPanel progressPanel = new JPanel();
        progressPanel.setLayout(new BoxLayout(progressPanel, BoxLayout.Y_AXIS));
        progressPanel.add(Box.createVerticalGlue());
        progressBarLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
        progressBarLabel.setForeground(new Color(8, 32, 128));
        progressPanel.add(progressBarLabel);
        progressPanel.add(Box.createRigidArea(new Dimension(10,10)));
        progressBar.setAlignmentX(Component.CENTER_ALIGNMENT);
        progressPanel.add(progressBar);
        progressPanel.add(Box.createRigidArea(new Dimension(10,30)));
        progressPanel.add(Box.createVerticalGlue());
        
        jPanelLoading.add(progressPanel);
        jPanelLoading.add(Box.createHorizontalGlue());        
        
        frame = new JFrame("TrainVis");
        JPanel jPanelGlob = new JPanel();
        final JTabbedPane tabPane = new JTabbedPane();
        tabPane.setPreferredSize(new Dimension(700, 650));
        
        menu = new Menu(tabPane, tablesZipFile);
        
        // fileChooser for tables
        JFileChooser fc = new JFileChooser();
        fc.setFileFilter( new FileFilter() {
            @Override public boolean accept( File f ) {
                return f.isDirectory() ||
                        f.getName().toLowerCase().endsWith( ".zip" );
            }
            @Override public String getDescription() {
                return "Texte";
            }
        } );

        final JFileChooser jFileChooser = new JFileChooser(".");
        jFileChooser.setFileFilter( new FileFilter() {
            @Override public boolean accept( File f ) {
                return f.isDirectory() ||
                        (f.getName().toLowerCase().endsWith( ".zip" ) &&
                        f.length() <= ZIPFILE_MAX_SIZE);
                
            }
            @Override public String getDescription() {
                return "Zip-Files";
            }
        } );
        
        menuBar = new JMenuBar();
        menuBar.setEnabled(false);
        File = new JMenu("File");
        menuBar.add(File);
        
        Load = new JMenuItem("Load Zip");
        Load.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                
                int returnVal = jFileChooser.showOpenDialog( null );
                
                if (returnVal == JFileChooser.APPROVE_OPTION) {
                    
                    if (jFileChooser.getSelectedFile().length() > ZIPFILE_MAX_SIZE) return;
                    menu.loadTables(jFileChooser.getSelectedFile().getPath());
                }
            }
        });
        File.add(Load);
                
        Exit = new JMenuItem("Exit");
        Exit.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                frame.dispose();
                System.exit(0);
            }
        });
        File.add(Exit);
        
                
        Direction = new JMenu("Direction");
        menuBar.add(Direction);
        fromto = new JCheckBoxMenuItem("");
        tofrom = new JCheckBoxMenuItem("");
        fromto.setSelected(true);
        tofrom.setSelected(true);
        fromto.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                if(fromto.isSelected() && tofrom.isSelected()) menu.setDirection(0);
                if(fromto.isSelected() && !tofrom.isSelected()) menu.setDirection(1);
                if(!fromto.isSelected() && tofrom.isSelected()) menu.setDirection(2);
                if(!fromto.isSelected() && !tofrom.isSelected()) {
                    tofrom.setSelected(true);
                    menu.setDirection(2);
                }
            }
        });
        
        tofrom.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                if(fromto.isSelected() && tofrom.isSelected()) menu.setDirection(0);
                if(fromto.isSelected() && !tofrom.isSelected()) menu.setDirection(1);
                if(!fromto.isSelected() && tofrom.isSelected()) menu.setDirection(2);
                if(!fromto.isSelected() && !tofrom.isSelected()) {
                    fromto.setSelected(true);
                    menu.setDirection(1);
                }
            }
        });
        
        Direction.add(fromto);
        Direction.add(tofrom);
        
        
        Sort = new JMenu("Sort");
        Sort.setEnabled(false);
        menuBar.add(Sort);
        sortUP = new JRadioButtonMenuItem("");
        sortDOWN = new JRadioButtonMenuItem("");
        
        sortUP.setSelected(true);
        
        sortUP.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sortDOWN.setSelected(false);
                menu.setSort(0);
            }
        });
        
        sortDOWN.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sortUP.setSelected(false);
                menu.setSort(1);
            }
        });
        Sort.add(sortUP);
        Sort.add(sortDOWN);
        
        Info = new JMenu("Info");
        menuBar.add(Info);
        About = new JMenuItem("About");
        Help = new JMenuItem("Help");
        Info.add(About);
        Info.add(Help);
        
        About.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                JOptionPane.showMessageDialog(frame, "" +
                        "<HTML>" +
                        "<b><font size=5>" +
                        "Train Visualization Tool" +
                        "</b></font><br><br>" +
                        "<font size = 4>" +
                        "Version 1.0<br><br>" +
                        "</font>" +
                        "Authors: &nbsp;Georg Prohaska (georgprohaska@gmx.at)<br>" +
                        "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Hansjorg Pirpamer (mail@hapi.it)<br> " +
                        "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Markus Martin (markus.martin@inode.at)<br>" +
                        " </HTML>");
            }
        });
        
        Help.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                HelpFrame.setVisible(true);
            }
        });
        
        frame.setJMenuBar(menuBar);        
        
        jPanelGlob.setLayout(new BorderLayout());
        jPanelGlob.add(menu, BorderLayout.NORTH);
        jPanelGlob.add(tabPane, BorderLayout.CENTER);
                
        frame.setContentPane(jPanelGlob);

        frame.pack();
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.setVisible(true);

        HelpFrame.setSize(HELP_WIDTH, HELP_HEIGHT);
        HelpFrame.setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
        
        
        // wird bei TabWechsel aufgerufen
        tabPane.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent evt) {
                
                try {
                    
                    // Men einschrnken
                    if (tabPane.getSelectedIndex() >= 0) {
                        
                        setMenu(tabPane, menu);
                        
                        // Nicht weitergeleitete nderungen weiterleiten
                        tabPane.getSelectedComponent().firePropertyChange("_FISHEYE_", -1, fishEye);
                        tabPane.getSelectedComponent().firePropertyChange("_STARTDATE_", oldstartDate, startDate);
                        tabPane.getSelectedComponent().firePropertyChange("_ENDDATE_", oldendDate, endDate);
                        tabPane.getSelectedComponent().firePropertyChange("_SETDIRECTION_", -1, direction);
                        tabPane.getSelectedComponent().firePropertyChange("_REPAINT_", -1, 1);
                    }
                } catch (Exception exc) { System.err.println("Can't process stateChanged"); }
            }
        });
        
        // alle PropertyChanges weiterleiten
        tabPane.addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                
                try {
                    
                    if (evt.getPropertyName() != null &&
                            evt.getOldValue() != null &&
                            evt.getNewValue() != null ) {
                        
                        menu.firePropertyChange(
                                evt.getPropertyName(),
                                Functions.object2Long(evt.getOldValue()),
                                Functions.object2Long(evt.getNewValue())
                                );
                        
                        boolean onlySelectedComponent = false;
                        
                        if (evt.getPropertyName().equals("_PUSHPROGRESS_")) {
                            
                            progressCounter += (Integer)evt.getNewValue();
                            progressBar.setValue(progressCounter);
                            onlySelectedComponent = true;
                        }
                        
                        if (evt.getPropertyName().equals("_FISHEYE_")) {
                            
                            fishEye = Functions.object2Long(evt.getNewValue());
                            onlySelectedComponent = true;
                        }
                        
                        if (evt.getPropertyName().equals("_SETDIRECTION_")) {
                            
                            direction = (Integer)evt.getNewValue();
                            onlySelectedComponent = true;
                        }
                        
                        if (evt.getPropertyName().equals("_STARTDATE_")) {
                            
                            startDate = Functions.object2Long(evt.getNewValue());
                            oldstartDate = Functions.object2Long(evt.getOldValue());
                            onlySelectedComponent = true;
                        }
                        
                        if (evt.getPropertyName().equals("_ENDDATE_")) {
                            
                            endDate = Functions.object2Long(evt.getNewValue());
                            oldendDate = Functions.object2Long(evt.getOldValue());
                            onlySelectedComponent = true;
                        }
                        
                        if (evt.getPropertyName().equals("_CHANGETABLE_")) {
                            
                            makeTabbedPane(tabPane, menu, tempDir);
                        }
                        
                        if (onlySelectedComponent)
                            tabPane.getSelectedComponent().firePropertyChange(
                                    evt.getPropertyName(),
                                    Functions.object2Long(evt.getOldValue()),
                                    Functions.object2Long(evt.getNewValue())
                                    );
                        
                        else {
                            
                            Component components[] = tabPane.getComponents();
                            
                            for (int i = 0; i < components.length; i++) {
                                components[i].firePropertyChange(
                                        evt.getPropertyName(),
                                        Functions.object2Long(evt.getOldValue()),
                                        Functions.object2Long(evt.getNewValue())
                                        );
                            }
                        }
                    }
                } catch (Exception exc) { System.err.println("Can't process propertyChange"); }
            }
        });
        
        makeTabbedPane(tabPane, menu, tempDir);
    }
}
