/*
 * Menu.java
 *
 * Created on 6. Juni 2007, 16:44
 *
 */

package zugvis;

import at.ac.tuwien.cs.timevis.FisheyeTimeScale;
import com.toedter.calendar.JCalendar;
import com.toedter.calendar.JDateChooser;
import java.awt.Component;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Vector;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JSeparator;
import javax.swing.JSlider;
import javax.swing.JTabbedPane;
import javax.swing.JToolBar;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.JCheckBox;
import prefuse.data.Table;

/**
 * This class represents the toolbar of the application. It contains all available components and their functionality.
 * @author hansjoerg
 */
public class Menu extends JToolBar{
    
    /**
     * Contains the visualisations
     */
    private JTabbedPane tabPane = null;
    
    /**
     * Button for zoom IN
     */    
    private JButton jButtonZoomPlus = new JButton();
    
    /**
     * Button for zoom OUT
     */    
    private JButton jButtonZoomMinus = new JButton();
    
    /**
     * Button for pan LEFT
     */    
    private JButton jButtonPanLeft = new JButton();
    
    /**
     * Button for pan RIGHT
     */    
    private JButton jButtonPanRight = new JButton();
    
    /**
     * Slider for fisheyeintensity
     */    
    private JSlider fisheyeSlider = new JSlider(0, 3000 , 0);
    
    /**
     * Flag if show stationnames
     */    
    private JCheckBox ShowStations = new JCheckBox("Show Stations");
    
    /**
     * Datechooser of startdate
     */    
    private JDateChooser jDateChooserStartDate = null;
    
    /**
     * Calendar of datechooser for startdate
     */    
    private JCalendar calendarStartDate = new JCalendar(); 
    
    /**
     * Datechooser of enddate
     */
    private JDateChooser jDateChooserEndDate = null;
    
    /**
     * Calendar of datechooser for enddate
     */
    private JCalendar calendarEndDate = new JCalendar(); 
    
    /**
     * Internal timescale for calculating start and enddate to send on visualizations
     */
    private FisheyeTimeScale timeScale;
        
    /**
     * Contains the zipfilename of the tables
     */
    private static Vector tablesZipFile = null;
    
    /**
     * Flag to show directions (both, from-to, to-from)
     */
    private int direction = 0;
    
    /**
     * Flag if the propertychange was fired external
     * (drag on visualization with leftbutton to pan or rightbutton to zoom for exampel)
     */
    private boolean externChangeStartDate = false;
    
    /**
     * Flag if the propertychange was fired external
     * (drag on visualization with leftbutton to pan or rightbutton to zoom for exampel)
     */
    private boolean externChangeEndDate = false;
    
    
    /**
     * Creates a new instance of Menu
     * @param tabPane The container of the visualzations (for propertyChange)
     * @param tablesZipFile Contains the filename of the zipfile with tables
     */
    public Menu(JTabbedPane tabPane, Vector tablesZipFile) {
        
        super("View");
        this.tabPane = tabPane;
        this.tablesZipFile = tablesZipFile;
        
        timeScale = new FisheyeTimeScale();
        timeScale.setAdjustDateRangeOnResize(false);
        timeScale.setDisplayWidth(700);
        
        addComponentListener(new ComponentAdapter() {
            public void componentResized(ComponentEvent e) {
                timeScale.setDisplayWidth(getWidth() - 1);
            }
        });
        
        setOrientation(HORIZONTAL);
       
        
        jButtonZoomPlus.setIcon(new ImageIcon(getClass().getClassLoader().getResource("zoomplus_on.gif")));
        jButtonZoomPlus.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                zoomTimeScale(1/1.2);
            }
        });
        add(jButtonZoomPlus);
        
        jButtonZoomMinus.setIcon(new ImageIcon(getClass().getClassLoader().getResource("zoomminus_on.gif")));
        jButtonZoomMinus.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                zoomTimeScale(1.2);
            }
        });
        add(jButtonZoomMinus);
        
        add(new JSeparator(VERTICAL));
        
        jButtonPanLeft.setIcon(new ImageIcon(getClass().getClassLoader().getResource("pan_left.gif")));
        jButtonPanLeft.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                panTimeScale(20);
            }
        });
        add(jButtonPanLeft);
        
        jButtonPanRight.setIcon(new ImageIcon(getClass().getClassLoader().getResource("pan_right.gif")));
        jButtonPanRight.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                panTimeScale(-20);
            }
        });
        add(jButtonPanRight);
        
        add(new JSeparator(VERTICAL));
        
        add(new JLabel(" FishEye "));
        
        fisheyeSlider.addChangeListener( new ChangeListener() {
            public void stateChanged( ChangeEvent evt ) {
                
                changeFisheye( (long) ((JSlider)evt.getSource()).getValue() );
            }
        } );
        add(fisheyeSlider);
        
        add(new JSeparator(VERTICAL));
        
        GregorianCalendar tempD = new GregorianCalendar(); 
        
        add(new JLabel(" from ")); 
        
        tempD.set(tempD.get(tempD.YEAR), tempD.get(tempD.MONTH), tempD.get(tempD.DAY_OF_MONTH), 0, 0, 0);
        jDateChooserStartDate = new JDateChooser(calendarStartDate, tempD.getTime(), null, null);
        setStartDate(tempD.getTime());        
        jDateChooserStartDate.addPropertyChangeListener(new java.beans.PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                changeStartDate(evt);
            }
        });        
        add(jDateChooserStartDate);
        
        add(new JLabel(" to "));
        
        tempD.set(tempD.get(tempD.YEAR), tempD.get(tempD.MONTH), tempD.get(tempD.DAY_OF_MONTH), 23, 59, 59);                
        jDateChooserEndDate = new JDateChooser(calendarEndDate, tempD.getTime(), null, null);
        setEndDate(tempD.getTime());        
        jDateChooserEndDate.addPropertyChangeListener(new java.beans.PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                changeEndDate(evt);
            }
        });
        add(jDateChooserEndDate);
        
        // Timecale can be changed outside from menu 
        // (drag on visualization with leftbutton to pan or rightbutton to zoom for exampel)
        addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                
                if (evt.getPropertyName().equals("_STARTDATE_")) {
                    
                    if(Functions.object2Long(evt.getOldValue()) == 1338) return;
                    externChangeStartDate = true;
                    setStartDate(new Date(Functions.object2Long(evt.getNewValue())));
                }
                
                if (evt.getPropertyName().equals("_ENDDATE_")) {
                    
                    if(Functions.object2Long(evt.getOldValue()) == 1338) return;
                    externChangeEndDate = true;
                    setEndDate(new Date(Functions.object2Long(evt.getNewValue())));
                }
            }
        });
        
        add(new JSeparator(VERTICAL));
        
        JButton jButtonFitTime = new JButton();
        jButtonFitTime.setText("FitTime");
        jButtonFitTime.setName("FITTIMEBUTTON");
        
        jButtonFitTime.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                FitTime();
            }
        });

        add(jButtonFitTime);
        jButtonFitTime.setBorderPainted(true);
        
        ShowStations.setSelected(true);
        ShowStations.setName("STATIONSCHECKBOX");
        
        add(new JSeparator(VERTICAL));
        add(ShowStations);
        
        ShowStations.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                ShowStations(ShowStations.isSelected());
            }
        });

        setAllEnabled(false);
    }
    
    
    /**
     * Set minimum selectable dates and maximum selectable dates of the calendars
     * @param dates Table of dates
     */
    public void setDates(Table dates) {
        
        if (dates != null) {
            
            calendarEndDate = Functions.setValidDates(calendarEndDate, dates);
            calendarStartDate = Functions.setValidDates(calendarStartDate, dates);
            
            calendarEndDate.setMinSelectableDate(calendarStartDate.getDate());
            calendarStartDate.setMaxSelectableDate(calendarEndDate.getDate());
        }
    }
    
    
    /**
     * Change the startdate of the menucalendars
     * @param newStartDate The new startdate
     */
    private void setStartDate(long newStartDate) { setStartDate(new Date(newStartDate)); }    
    
    /**
     * Change the startdate of the menucalendars
     * @param newStartDate The new startdate
     */
    private void setStartDate(Date newStartDate) {
        
        //System.out.println("Start: " + newStartDate);
        calendarEndDate.setMinSelectableDate(newStartDate);

        
        calendarStartDate.setDate(newStartDate);
        jDateChooserStartDate.setDate(newStartDate);
        timeScale.setStartDate(newStartDate.getTime());
    }
    
    
    /**
     * Change the endate of the menucalendars
     * @param newEndDate The new enddate
     */
    private void setEndDate(long newEndDate) { setEndDate(new Date(newEndDate)); }
        
    /**
     * Change the endate of the menucalendars
     * @param newEndDate The new enddate
     */
    private void setEndDate(Date newEndDate) {
        
        calendarStartDate.setMaxSelectableDate(newEndDate);
        
        calendarEndDate.setDate(newEndDate);
        jDateChooserEndDate.setDate(newEndDate);
        timeScale.setEndDate(newEndDate.getTime());
    }
    
    
    /**
     * With this method one can set the enabled-flag of all components of the toolbar.
     * @param enabled enabled Falg for enabling or disabling the menucomponents
     */
    public void setAllEnabled(boolean enabled) {
        
        Component components[] = getComponents();
        
        for (int i = 0; i < components.length; i++) {

            if (components[i].getName() != null &&
                    components[i].getName().equals("FILECHOOSER_BUTTON"))                 
                components[i].setEnabled(true);            
            else components[i].setEnabled(enabled);           
        }
    }
    
    
    /**
     * Sets the enabled flag of all components of the toolbar that are used to manipulate the timescale.
     * @param enabled Falg for enabling or disabling the timemenucomponents
     */
    public void setTimeControlsEnabled(boolean enabled) {
        
        jButtonZoomPlus.setEnabled(enabled);
        jButtonZoomMinus.setEnabled(enabled);
        jButtonPanLeft.setEnabled(enabled);
        jButtonPanRight.setEnabled(enabled);
        fisheyeSlider.setEnabled(enabled);
    }
    
    
    /**
     * Sets the enabled flag of a specific component.
     * @param enabled Falg for enabling or disabling the specific component
     * @param componentName The name of the component.
     */
    public void setComponetEnabled(String componentName, boolean enabled) {
        
        Component components[] = getComponents();
        
        for (int i = 0; i < components.length; i++) {

            if (components[i].getName() != null &&
                    components[i].getName().equals(componentName))                 
                components[i].setEnabled(enabled);
        }    
    }
    
    
    /**
     * Fires a _SETDIRECTION_ event to the tabpane.
     * @param direction directionmode, 0 = both directions, 1 = only "from-to"-connections, 2 = only "to-from"-connections
     */
    public void setDirection(int direction) {
        
        tabPane.firePropertyChange("_SETDIRECTION_", -1, direction);
    }
     
    
    /**
     * Fires a _SETSORT_ event to the tabpane.
     * @param sort Sortdirection, 0 = UP, 1 = DOWN
     */
    public void setSort(int sort) {
        
        tabPane.firePropertyChange("_SETSORT_", -1, sort);
        
    }
    
    
    /**
     * Fires a _FITTIME_ event to the tabpane.
     */
    private void FitTime(){
        
        tabPane.firePropertyChange("_FITTIME_", -1, 1);        
    } 
    
    
    /**
     * Fires a _SHOWSTATIONS_ event to the tabpane.
     * @param show If true, stationnames where shown
     */
    private void ShowStations(boolean show){
        
        tabPane.firePropertyChange("_SHOWSTATIONS_", -1, show ? 1 : 0);
    }
     
    
    /**
     * Loads a new Zip file by fireing a _CHANGETABLE_ event.
     * @param fileName Filename of zipfile
     */
    public void loadTables(String fileName) {
        
        tablesZipFile.add(0, fileName);
        tabPane.firePropertyChange("_CHANGETABLE_", 0, 1);
    }    
    
    
    /**
     * Change de zoomlevel of the hidden menutimescale 
     * anf fires a _STARTDATE_ and _ENDDATE_ event to the tabpane
     * @param zoom Zoomfactor
     */
    private void zoomTimeScale(double zoom) {
        
        Date oldStartDate = new Date(timeScale.getStartDate());
        Date oldEndDate = new Date(timeScale.getEndDate());
        
        timeScale.zoom( zoom );
        
        Date newStartDate = new Date(timeScale.getStartDate());
        Date newEndDate = new Date(timeScale.getEndDate());
        
        setStartDate(newStartDate);
        tabPane.firePropertyChange("_STARTDATE_", 1338, newStartDate.getTime());
        
        setEndDate(newEndDate);
        tabPane.firePropertyChange("_ENDDATE_", 1338, newEndDate.getTime());
    }
    
    
    /**
     * Pan the hidden menutimescale and fires 
     * a _STARTDATE_ and _ENDDATE_ event to the tabpane
     * @param pan Panfactor
     */
    private void panTimeScale(long pan) {
        
        long oldStartDate = timeScale.getStartDate();
        long oldEndDate = timeScale.getEndDate();
        
        long diffMillis = timeScale.getMillisPerPixel() * pan;
        
        long newStartDate = oldStartDate - diffMillis;
        long newEndDate = oldEndDate - diffMillis;
        
        setStartDate(newStartDate);
        tabPane.firePropertyChange("_STARTDATE_", 1338, newStartDate);
        
        setEndDate(newEndDate);
        tabPane.firePropertyChange("_ENDDATE_", 1338, newEndDate);
    }
    
    
    /**
     * Fires a _FISHEYE_ event to the tabpane.
     * @param fisheye Fisheyelevel
     */
    private void changeFisheye(long fisheye) {
        
        tabPane.firePropertyChange("_FISHEYE_", 1, fisheye);
    }    
    
    
    /**
     * Changes the startdate and fires a _STARTDATE_ event to the other components
     * @param evt Event
     */
    private void changeStartDate(PropertyChangeEvent evt) {
        
        if (evt.getPropertyName().equals("date")) {
            
            Date oldDate = (Date)evt.getOldValue();
            GregorianCalendar newDate = new GregorianCalendar();
            newDate.setTime((Date)evt.getNewValue());
            newDate.set(newDate.get(newDate.YEAR), newDate.get(newDate.MONTH), newDate.get(newDate.DAY_OF_MONTH), 0, 0, 0);
            
            
            // prevent a too small range
            if (Math.abs(newDate.getTime().compareTo(calendarStartDate.getMaxSelectableDate())) > 1000*60*60) return;
            
            boolean temp = externChangeStartDate;
            setStartDate(newDate.getTime());
            externChangeStartDate = temp;
            
           
            if (!externChangeStartDate) tabPane.firePropertyChange("_STARTDATE_", 1338, newDate.getTimeInMillis());
            else externChangeStartDate = false;
        }
    }
    
    
    /**
     * Changes the enddate and fires a _ENDDATE_ event to the other components
     * @param evt Event
     */
    private void changeEndDate(PropertyChangeEvent evt) {
        
        if (evt.getPropertyName().equals("date")) {
            
            Date oldDate = (Date)evt.getOldValue();
            GregorianCalendar newDate = new GregorianCalendar();
            newDate.setTime((Date)evt.getNewValue());
            newDate.set(newDate.get(newDate.YEAR), newDate.get(newDate.MONTH), newDate.get(newDate.DAY_OF_MONTH), 23, 59, 59);
            
            
            // prevent a too small range
            if (Math.abs(newDate.getTime().compareTo(calendarStartDate.getMinSelectableDate())) > 1000*60*60) return;
            
            boolean temp = externChangeEndDate;
            setEndDate(newDate.getTime());
            externChangeEndDate = temp;
            
            
            if (!externChangeEndDate) tabPane.firePropertyChange("_ENDDATE_", 1338, newDate.getTimeInMillis());
            else externChangeEndDate = false;
        }
    }
}





